import gymnasium as gym

from metadrive.component.vehicle.PID_controller import PIDController
from metadrive.engine.engine_utils import get_global_config
from metadrive.policy.env_input_policy import EnvInputPolicy
from metadrive.utils.math import wrap_to_pi


class AgentLaneChangePolicy(EnvInputPolicy):
    def __init__(self, obj, seed):
        # Since control object may change
        super(AgentLaneChangePolicy, self).__init__(obj, seed)
        self.discrete_action = self.engine.global_config["discrete_action"]
        assert self.discrete_action, "Must set discrete_action=True for using this control policy"
        self.use_multi_discrete = self.engine.global_config["use_multi_discrete"]
        self.steering_unit = 1.0
        self.throttle_unit = 2.0 / (
            self.engine.global_config["discrete_throttle_dim"] - 1
        )  # for discrete actions space
        self.discrete_steering_dim = 3  # only left or right
        self.discrete_throttle_dim = self.engine.global_config["discrete_throttle_dim"]

        self.heading_pid = PIDController(1.7, 0.01, 3.5)
        self.lateral_pid = PIDController(0.3, .002, 0.05)

    def act(self, agent_id):
        action = super(AgentLaneChangePolicy, self).act(agent_id)
        steering = action[0]
        throttle = action[1]
        current_lane = self.control_object.navigation.current_lane
        if steering == 0.0:
            target_lane = current_lane
        elif steering == 1.0:
            target_lane = self.control_object.navigation.current_ref_lanes[max(current_lane.index[-1] - 1, 0)]
        elif steering == -1.0:
            lane_num = len(self.control_object.navigation.current_ref_lanes)
            target_lane = self.control_object.navigation.current_ref_lanes[
                min(current_lane.index[-1] + 1, lane_num - 1)]
        else:
            raise ValueError("Steering Error, can only be in [-1, 0, 1]")
        action = [self.steering_control(target_lane), throttle]
        self.action_info["action"] = action
        return action

    @classmethod
    def get_input_space(cls):
        """
       The Input space is a class attribute
       """
        engine_global_config = get_global_config()
        extra_action_dim = engine_global_config["vehicle_config"]["extra_action_dim"]
        discrete_action = engine_global_config["discrete_action"]
        assert discrete_action
        discrete_throttle_dim = engine_global_config["discrete_throttle_dim"]
        use_multi_discrete = engine_global_config["use_multi_discrete"]
        discrete_steering_dim = 3

        if use_multi_discrete:
            return gym.spaces.MultiDiscrete([discrete_steering_dim, discrete_throttle_dim])
        else:
            return gym.spaces.Discrete(discrete_steering_dim * discrete_throttle_dim)

    def steering_control(self, target_lane) -> float:
        # heading control following a lateral distance control
        ego_vehicle = self.control_object
        long, lat = target_lane.local_coordinates(ego_vehicle.position)
        lane_heading = target_lane.heading_theta_at(long + 1)
        v_heading = ego_vehicle.heading_theta
        steering = self.heading_pid.get_result(-wrap_to_pi(lane_heading - v_heading))
        steering += self.lateral_pid.get_result(-lat)
        return float(steering)



"""
This script demonstrates how to use the environment where traffic and road map are loaded from Waymo dataset.
"""
import argparse
import random

from metadrive.constants import HELP_MESSAGE
from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv


class DemoWaymoEnv(WaymoEnv):
    def reset(self, seed=None):
        if self.engine is not None:
            seeds = [i for i in range(self.config["num_scenarios"])]
            seeds.remove(self.current_seed)
            seed = random.choice(seeds)
        return super(DemoWaymoEnv, self).reset(seed=seed)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--reactive_traffic", action="store_true")
    parser.add_argument("--top_down", action="store_true")
    args = parser.parse_args()
    extra_args = dict(film_size=(800, 800)) if args.top_down else {}
    asset_path = AssetLoader.asset_path
    print(HELP_MESSAGE)
    try:
        env = DemoWaymoEnv(
            {
                "manual_control": True,
                "reactive_traffic": True if args.reactive_traffic else False,
                "use_render": True if not args.top_down else False,
                "data_directory": AssetLoader.file_path(asset_path, "waymo", return_raw_style=False),
                "num_scenarios": 3
            }
        )
        o, _ = env.reset()

        for i in range(1, 100000):
            o, r, tm, tc, info = env.step([1.0, 0.])
            env.render(
                mode="top_down" if args.top_down else None,
                text={
                    "Quit": "ESC",
                    "Switch perspective": "Q or B",
                    "Reset Episode": "R",
                    "Keyboard Control": "W,A,S,D",
                },
                **extra_args
            )
            if tm or tc:
                env.reset()
    except Exception as e:
        raise e
    finally:
        env.close()



import numpy as np
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.component.sensors.mini_map import MiniMap
from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive.component.sensors.vehicle_panel import VehiclePanel

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

if __name__ == "__main__":
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 10,
            "traffic_density": 0.2,
            "traffic_mode": "hybrid",
            "start_seed": 22,
            "debug": True,
            "cull_scene": False,
            "manual_control": True,
            "use_render": True,
            "decision_repeat": 5,
            "interface_panel": [MiniMap, VehiclePanel, RGBCamera],
            "need_inverse_traffic": False,
            "rgb_clip": True,
            "map": "SSS",
            # "agent_policy": IDMPolicy,
            "random_traffic": False,
            "random_lane_width": True,
            "random_agent_model": True,
            "driving_reward": 1.0,
            "force_destroy": False,
            "show_interface": False,
            "vehicle_config": {
                "enable_reverse": False,
            },
        }
    )
    import time

    start = time.time()
    o, _ = env.reset()

    def get_v_path():
        return BaseVehicle.model_collection[env.vehicle.path[0]]

    def add_x():
        model = get_v_path()
        model.setX(model.getX() + 0.1)
        # print(model.getPos())

    def decrease_x():
        model = get_v_path()
        model.setX(model.getX() - 0.1)
        # print(model.getPos())

    def add_y():
        model = get_v_path()
        model.setY(model.getY() + 0.1)
        # print(model.getPos())

    def decrease_y():
        model = get_v_path()
        model.setY(model.getY() - 0.1)
        # print(model.getPos())

    env.engine.accept("i", add_x)
    env.engine.accept("k", decrease_x)
    env.engine.accept("j", add_y)
    env.engine.accept("l", decrease_y)

    for s in range(1, 10000):
        o, r, tm, tc, i = env.step([0, 0])
        env.render(
            text={
                "heading_diff": env.vehicle.heading_diff(env.vehicle.lane),
                "lane_width": env.vehicle.lane.width,
                "lateral": env.vehicle.lane.local_coordinates(env.vehicle.position),
                "current_seed": env.current_seed
            }
        )



import math
from typing import Tuple

import numpy as np
from scipy.interpolate import interp1d

number_pos_inf = float("inf")
number_neg_inf = float("-inf")


def safe_clip(array, min_val, max_val):
    array = np.nan_to_num(array.astype(np.float64), copy=False, nan=0.0, posinf=max_val, neginf=min_val)
    return np.clip(array, min_val, max_val).astype(np.float64)


def safe_clip_for_small_array(array, min_val, max_val):
    array = list(array)
    for i in range(len(array)):
        if math.isnan(array[i]):
            array[i] = 0.0
        elif array[i] == number_pos_inf:
            array[i] = max_val
        elif array[i] == number_neg_inf:
            array[i] = min_val
        array[i] = clip(array[i], min_val, max_val)
    return array


def wrap_to_pi(x: float) -> float:
    angles = x
    angles %= 2 * np.pi
    angles -= 2 * np.pi * (angles > np.pi)
    return angles


def get_vertical_vector(vector: np.array):
    length = norm(vector[0], vector[1])
    # return (vector[1] / length, -vector[0] / length), (-vector[1] / length, vector[0] / length)
    return (-vector[1] / length, vector[0] / length), (vector[1] / length, -vector[0] / length)


def norm(x, y):
    return math.sqrt(x**2 + y**2)


def clip(a, low, high):
    return min(max(a, low), high)


def point_distance(x, y):
    return norm(x[0] - y[0], x[1] - y[1])


def panda_vector(position_x, position_y, z=0.0):
    return position_x, position_y, z


def distance_greater(vec1, vec2, length):
    """Return whether the distance between two vectors is greater than the given length."""
    return ((vec1[0] - vec2[0])**2 + (vec1[1] - vec2[1])**2) > length**2


def mph_to_kmh(speed_in_mph: float):
    speed_in_kmh = speed_in_mph * 1.609344
    return speed_in_kmh


def get_laser_end(lidar_range, perceive_distance, laser_index, heading_theta, vehicle_position_x, vehicle_position_y):
    angle = lidar_range[laser_index] + heading_theta
    return (
        perceive_distance * math.cos(angle) + vehicle_position_x,
        perceive_distance * math.sin(angle) + vehicle_position_y
    )


def dot(a, b):
    return a[0] * b[0] + a[1] * b[1]


def dot3(a, b):
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def do_every(duration: float, timer: float) -> bool:
    return duration < timer


def not_zero(x: float, eps: float = 1e-2) -> float:
    if abs(x) > eps:
        return x
    elif x > 0:
        return eps
    else:
        return -eps


def rotated_rectangles_intersect(rect1: Tuple, rect2: Tuple) -> bool:
    """
    Do two rotated rectangles intersect?

    :param rect1: (center, length, width, angle)
    :param rect2: (center, length, width, angle)
    :return: do they?
    """
    return has_corner_inside(rect1, rect2) or has_corner_inside(rect2, rect1)


def point_in_rectangle(point, rect_min, rect_max) -> bool:
    """
    Check if a point is inside a rectangle

    :param point: a point (x, y)
    :param rect_min: x_min, y_min
    :param rect_max: x_max, y_max
    """
    return rect_min[0] <= point[0] <= rect_max[0] and rect_min[1] <= point[1] <= rect_max[1]


def point_in_rotated_rectangle(point: np.ndarray, center: np.ndarray, length: float, width: float, angle: float) \
        -> bool:
    """
    Check if a point is inside a rotated rectangle

    :param point: a point
    :param center: rectangle center
    :param length: rectangle length
    :param width: rectangle width
    :param angle: rectangle angle [rad]
    :return: is the point inside the rectangle
    """
    c, s = math.cos(angle), math.sin(angle)
    r = np.array([[c, -s], [s, c]])
    ru = r.dot(point - center)
    return point_in_rectangle(ru, (-length / 2, -width / 2), (length / 2, width / 2))


def has_corner_inside(rect1: Tuple, rect2: Tuple) -> bool:
    """
    Check if rect1 has a corner inside rect2

    :param rect1: (center, length, width, angle)
    :param rect2: (center, length, width, angle)
    """
    (c1, l1, w1, a1) = rect1
    (c2, l2, w2, a2) = rect2
    c1 = np.array(c1)
    l1v = np.array([l1 / 2, 0])
    w1v = np.array([0, w1 / 2])
    r1_points = np.array([[0, 0], -l1v, l1v, -w1v, w1v, -l1v - w1v, -l1v + w1v, +l1v - w1v, +l1v + w1v])
    c, s = math.cos(a1), math.sin(a1)
    r = np.array([[c, -s], [s, c]])
    rotated_r1_points = r.dot(r1_points.transpose()).transpose()
    return any([point_in_rotated_rectangle(c1 + np.squeeze(p), c2, l2, w2, a2) for p in rotated_r1_points])


def get_points_bounding_box(line_points):
    """
    Get bounding box from several points
    :param line_points: Key points on lines
    :return: bounding box
    """
    new_line_points = np.array(line_points)
    new_x_max = new_line_points[:, 0].max()
    new_x_min = new_line_points[:, 0].min()
    new_y_max = new_line_points[:, 1].max()
    new_y_min = new_line_points[:, 1].min()
    return new_x_max, new_x_min, new_y_max, new_y_min


def get_boxes_bounding_box(boxes):
    """
    Get a max bounding box from sveral small bound boxes
    :param boxes: List of other bounding box
    :return: Max bounding box
    """
    res_x_max = -np.inf
    res_x_min = np.inf
    res_y_min = np.inf
    res_y_max = -np.inf
    for x_max, x_min, y_max, y_min in boxes:
        res_x_max = max(res_x_max, x_max)
        res_x_min = min(res_x_min, x_min)
        res_y_max = max(res_y_max, y_max)
        res_y_min = min(res_y_min, y_min)
    return res_x_max, res_x_min, res_y_max, res_y_min


class Vector(tuple):
    """
    Avoid using this data structure!
    """
    def __sub__(self, other):
        return Vector((self[0] - other[0], self[1] - other[1]))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        if isinstance(other, float) or np.isscalar(other):
            return Vector((self[0] * other, self[1] * other))
        else:
            return Vector((self[0] * other[0], self[1] * other[1]))

    def __add__(self, other):
        if isinstance(other, float) or np.isscalar(other):
            return Vector((self[0] + other, self[1] + other))
        else:
            return Vector((self[0] + other[0], self[1] + other[1]))

    def __truediv__(self, other):
        if isinstance(other, float) or np.isscalar(other):
            ret = Vector((self[0] / other, self[1] / other))
            return ret
        raise ValueError()

    def tolist(self):
        return list(self)

    def __rsub__(self, other):
        return Vector(other) - self

    def __neg__(self):
        return Vector((-self[0], -self[1]))

    def dot(self, other):
        return self[0] * other[0] + self[1] * other[1]


def compute_angular_velocity(initial_heading, final_heading, dt):
    """
    Calculate the angular velocity between two headings given in radians.

    Parameters:
    initial_heading (float): The initial heading in radians.
    final_heading (float): The final heading in radians.
    dt (float): The time interval between the two headings in seconds.

    Returns:
    float: The angular velocity in radians per second.
    """

    # Calculate the difference in headings
    delta_heading = final_heading - initial_heading

    # Adjust the delta_heading to be in the range (-π, π]
    delta_heading = (delta_heading + math.pi) % (2 * math.pi) - math.pi

    # Compute the angular velocity
    angular_vel = delta_heading / dt

    return angular_vel


def get_polyline_length(points_array):
    diff = np.diff(points_array, axis=0)
    squared_diff = diff**2
    squared_diff_sum = np.sum(squared_diff, axis=1)
    distances = np.sqrt(squared_diff_sum)
    return np.sum(distances)


def resample_polyline(points, target_distance):
    # Calculate the cumulative distance along the original polyline
    distances = np.cumsum(np.sqrt(np.sum(np.diff(points, axis=0)**2, axis=1)))
    distances = np.insert(distances, 0, 0., axis=0)

    # Create a linearly spaced array of distances for the resampled polyline
    resampled_distances = np.arange(0, distances[-1], target_distance)

    # Interpolate the points along the resampled distances
    resampled_points = interp1d(distances, points, axis=0)(resampled_distances)

    return resampled_points



import copy

import numpy as np
from math import floor
from panda3d.bullet import BulletBoxShape, BulletGhostNode
from panda3d.core import Vec3

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import MetaDriveType
from metadrive.constants import CollisionGroup
from metadrive.engine.engine_utils import get_engine
from metadrive.manager.base_manager import BaseManager
from metadrive.utils import Config
from metadrive.utils.coordinates_shift import panda_vector, panda_heading
from metadrive.utils.pg.utils import rect_region_detection


class SpawnManager(BaseManager):
    """
    This class maintain a list of possible spawn places/destination for MARL respawn
    """
    # it needs to fill the config at first
    PRIORITY = 1

    REGION_DETECT_HEIGHT = 10
    RESPAWN_REGION_LONGITUDE = 8.
    RESPAWN_REGION_LATERAL = 3.
    MAX_VEHICLE_LENGTH = BaseVehicle.MAX_LENGTH
    MAX_VEHICLE_WIDTH = BaseVehicle.MAX_WIDTH

    # lazy init now
    initialized = False

    def __init__(self):
        # Lazy init~
        super(SpawnManager, self).__init__()
        self.initialized = True
        self.num_agents = self.engine.global_config["num_agents"]
        self.exit_length = (self.engine.global_config["map_config"]["exit_length"] - FirstPGBlock.ENTRANCE_LENGTH)
        assert self.exit_length >= self.RESPAWN_REGION_LONGITUDE, (
            "The exist length {} should greater than minimal longitude interval {}.".format(
                self.exit_length, self.RESPAWN_REGION_LONGITUDE
            )
        )
        self.lane_num = self.engine.global_config["map_config"]["lane_num"]
        self.spawn_roads = []
        self.safe_spawn_places = {}
        self.need_update_spawn_places = True
        self.spawn_places_used = []  # reset every step

        target_vehicle_configs = copy.copy(self.engine.global_config["target_vehicle_configs"])
        self._init_target_vehicle_configs = target_vehicle_configs

        spawn_roads = self.engine.global_config["spawn_roads"]
        target_vehicle_configs, safe_spawn_places = self._auto_fill_spawn_roads_randomly(spawn_roads)
        self.available_target_vehicle_configs = target_vehicle_configs
        self.safe_spawn_places = {place["identifier"]: place for place in safe_spawn_places}
        self.spawn_roads = spawn_roads
        self.need_update_spawn_places = True

    @staticmethod
    def get_not_randomize_vehicle_configs(configs):
        raise DeprecationWarning
        ret = {}
        for id, config in configs.items():
            if config["not_randomize"]:
                ret[id] = config
        return ret

    def reset(self):
        # random assign spawn points
        num_agents = self.num_agents if self.num_agents is not None else len(self.available_target_vehicle_configs)
        assert len(self.available_target_vehicle_configs) > 0

        if num_agents == -1:  # Infinite number of agents
            target_agents = list(range(len(self.available_target_vehicle_configs)))
        else:
            target_agents = self.np_random.choice(
                [i for i in range(len(self.available_target_vehicle_configs))], num_agents, replace=False
            )

        # set the spawn road
        ret = {}
        if len(target_agents) > 1:
            for real_idx, idx in enumerate(target_agents):
                v_config = self.available_target_vehicle_configs[idx]["config"]
                v_config = self._randomize_position_in_slot(v_config)
                ret["agent{}".format(real_idx)] = v_config
        else:
            ret["agent0"] = self._randomize_position_in_slot(self.available_target_vehicle_configs[0]["config"])

        # set the destination/spawn point and update target_v config
        target_vehicle_configs = {}
        for agent_id, config in ret.items():
            init_config = copy.deepcopy(self._init_target_vehicle_configs[agent_id])
            if not init_config.get("_specified_spawn_lane", False):
                init_config.update(config)
            config = init_config
            if not config.get("destination", False) or config["destination"] is None:
                config = self.update_destination_for(agent_id, config)
            target_vehicle_configs[agent_id] = config

        self.engine.global_config["target_vehicle_configs"] = copy.deepcopy(target_vehicle_configs)

    @staticmethod
    def max_capacity(spawn_roads, exit_length, lane_num):
        exit_length -= FirstPGBlock.ENTRANCE_LENGTH
        assert len(spawn_roads) > 0
        num_slots = int(floor(exit_length / SpawnManager.RESPAWN_REGION_LONGITUDE))
        assert num_slots > 0, "The exist length {} should greater than minimal longitude interval {}.".format(
            exit_length, SpawnManager.RESPAWN_REGION_LONGITUDE + FirstPGBlock.ENTRANCE_LENGTH
        )
        return lane_num * len(spawn_roads) * num_slots

    def _auto_fill_spawn_roads_randomly(self, spawn_roads):
        """It is used for shuffling the config"""

        num_slots = int(floor(self.exit_length / SpawnManager.RESPAWN_REGION_LONGITUDE))
        interval = self.exit_length / num_slots
        self._longitude_spawn_interval = interval
        if self.num_agents is not None:
            assert self.num_agents > 0 or self.num_agents == -1
            assert self.num_agents <= self.max_capacity(
                spawn_roads, self.exit_length + FirstPGBlock.ENTRANCE_LENGTH, self.lane_num
            ), (
                "Too many agents! We only accepet {} agents, but you have {} agents!".format(
                    self.lane_num * len(spawn_roads) * num_slots, self.num_agents
                )
            )

        # We can spawn agents in the middle of road at the initial time, but when some vehicles need to be respawn,
        # then we have to set it to the farthest places to ensure safety (otherwise the new vehicles may suddenly
        # appear at the middle of the road!)
        target_vehicle_configs = []
        safe_spawn_places = []
        for i, road in enumerate(spawn_roads):
            for lane_idx in range(self.lane_num):
                for j in range(num_slots):
                    long = 1 / 2 * self.RESPAWN_REGION_LONGITUDE + j * self.RESPAWN_REGION_LONGITUDE
                    lane_tuple = road.lane_index(lane_idx)  # like (>>>, 1C0_0_, 1) and so on.
                    target_vehicle_configs.append(
                        Config(
                            dict(
                                identifier="|".join((str(s) for s in lane_tuple + (j, ))),
                                config={
                                    "spawn_lane_index": lane_tuple,
                                    "spawn_longitude": long,
                                    "spawn_lateral": 0
                                },
                            ),
                            unchangeable=True
                        )
                    )  # lock the spawn positions
                    if j == 0:
                        safe_spawn_places.append(copy.deepcopy(target_vehicle_configs[-1]))
        return target_vehicle_configs, safe_spawn_places

    def step(self):
        self.spawn_places_used = []

    def get_available_respawn_places(self, map, randomize=False):
        """
        In each episode, we allow the vehicles to respawn at the start of road, randomize will give vehicles a random
        position in the respawn region
        """
        engine = get_engine()
        ret = {}
        for bid, bp in self.safe_spawn_places.items():
            if bid in self.spawn_places_used:
                continue
            # save time calculate once
            if not bp.get("spawn_point_position", False):
                lane = map.road_network.get_lane(bp["config"]["spawn_lane_index"])
                assert isinstance(lane, StraightLane), "Now we don't support respawn on circular lane"
                long = self.RESPAWN_REGION_LONGITUDE / 2
                spawn_point_position = lane.position(longitudinal=long, lateral=0)
                bp.force_update(
                    {
                        "spawn_point_heading": np.rad2deg(lane.heading_theta_at(long)),
                        "spawn_point_position": (spawn_point_position[0], spawn_point_position[1])
                    }
                )

            spawn_point_position = bp["spawn_point_position"]
            lane_heading = bp["spawn_point_heading"]
            result = rect_region_detection(
                engine, spawn_point_position, lane_heading, self.RESPAWN_REGION_LONGITUDE, self.RESPAWN_REGION_LATERAL,
                CollisionGroup.Vehicle
            )
            if (engine.global_config["debug"] or engine.global_config["debug_physics_world"]) \
                    and bp.get("need_debug", True):
                shape = BulletBoxShape(Vec3(self.RESPAWN_REGION_LONGITUDE / 2, self.RESPAWN_REGION_LATERAL / 2, 1))
                vis_body = engine.render.attach_new_node(BulletGhostNode("debug"))
                vis_body.node().addShape(shape)
                vis_body.setH(panda_heading(lane_heading))
                vis_body.setPos(panda_vector(spawn_point_position, z=2))
                engine.physics_world.dynamic_world.attach(vis_body.node())
                vis_body.node().setIntoCollideMask(CollisionGroup.AllOff)
                bp.force_set("need_debug", False)

            if not result.hasHit() or result.node.getName() != MetaDriveType.VEHICLE:
                new_bp = copy.deepcopy(bp).get_dict()
                if randomize:
                    new_bp["config"] = self._randomize_position_in_slot(new_bp["config"])
                ret[bid] = new_bp
                self.spawn_places_used.append(bid)
        return ret

    def _randomize_position_in_slot(self, target_vehicle_config):
        vehicle_config = copy.deepcopy(target_vehicle_config)
        long = self.RESPAWN_REGION_LONGITUDE - self.MAX_VEHICLE_LENGTH
        lat = self.RESPAWN_REGION_LATERAL - self.MAX_VEHICLE_WIDTH
        vehicle_config["spawn_longitude"] += self.np_random.uniform(-long / 2, long / 2)
        vehicle_config["spawn_lateral"] += self.np_random.uniform(-lat / 2, lat / 2)
        return vehicle_config

    def seed(self, random_seed):
        """this class is used to ranomly choose the spawn places, which will not be controlled by any seed"""
        if self.engine.global_config["force_seed_spawn_manager"]:
            super(SpawnManager, self).seed(random_seed)

    def update_destination_for(self, agent_id, vehicle_config):
        """
        Choose a destination for agent
        """
        return vehicle_config



"""

Script to convert spi1d files to luts

"""

from __future__ import print_function

import os
from panda3d.core import PNMImage

for f in os.listdir("film_luts_raw"):
    if not f.endswith(".spi1d"):
        continue

    input_file = "film_luts_raw/" + f

    print("Processing", input_file)
    output_name = input_file.replace("\\", "/").split("/")[-1].replace(".spi1d", "")

    with open(input_file, "r") as handle:
        content = [i.strip() for i in handle.readlines()]

    header = [
        "Version 1",
        "From 0.0 1.0",
        "Length 1024",
        "Components 3",
        "{",
    ]

    for i, line in enumerate(header):
        if content[i] != line:
            print("Invalid header! Expected", line, "but got", content[i])
            break
    else:

        content = content[len(header):-1]

        values = [[], [], []]

        def srgb_to_linear(channel):
            if channel <= 0.04045:
                return channel / 12.92
            else:
                return pow((channel + 0.055) / (1.0 + 0.055), 2.4)

        for i in xrange(1024):
            line = [float(k) for k in content[i].split()]

            for k in xrange(3):
                values[k].append(srgb_to_linear(line[k]))

        def to_linear(v):
            return float(v) / float(64 - 1)

        def to_linear_inv(v):
            return 1 - to_linear(v)

        def lookup_value(v, values):
            return values[int(v * (len(values) - 1))]

        # Generate lut
        img = PNMImage(64 * 8, 64 * 8, 3, 2**16 - 1)
        for r in xrange(64):
            for g in xrange(64):
                for b in xrange(64):
                    slice_offset_x = (b % 8) * 64
                    slice_offset_y = (b // 8) * 64

                    fr, fg, fb = to_linear(r), to_linear_inv(g), to_linear(b)

                    fr = lookup_value(fr, values[0])
                    fg = lookup_value(fg, values[1])
                    fb = lookup_value(fb, values[2])

                    img.set_xel(r + slice_offset_x, g + slice_offset_y, fr, fg, fb)

        img.write("film_luts/" + output_name + ".png")



from collections import defaultdict
from metadrive.component.algorithm.blocks_prob_dist import PGBlockDistConfig
from typing import Union

from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import engine_initialized
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.manager.traffic_manager import PGTrafficManager
from metadrive.manager.traffic_manager import TrafficMode
from metadrive.manager.waymo_data_manager import WaymoDataManager
from metadrive.manager.waymo_map_manager import WaymoMapManager
from metadrive.manager.waymo_traffic_manager import WaymoTrafficManager
from metadrive.utils import get_np_random

MIX_WAYMO_PG_ENV_CONFIG = dict(
    # ===== Waymo Map Config =====
    data_directory=AssetLoader.file_path("waymo", return_raw_style=False),
    start_scenario_index=0,
    num_scenarios=50,
    store_map=True,
    store_map_buffer_size=200,

    # ===== Waymo Traffic Config =====
    no_traffic=False,
    reactive_traffic=False,

    # ===== PG Map config =====
    start_seed=0,
    block_dist_config=PGBlockDistConfig,

    # ===== PG Map Config =====
    block_num=1,  # block_num
    random_lane_width=False,
    random_lane_num=False,

    # ===== PG Traffic =====
    traffic_density=0.2,
    need_inverse_traffic=False,
    traffic_mode=TrafficMode.Hybrid,  # "Respawn", "Trigger"
    random_traffic=False,  # Traffic is randomized at default.
    # this will update the vehicle_config and set to traffic
    traffic_vehicle_config=dict(
        show_navi_mark=False,
        show_dest_mark=False,
        enable_reverse=False,
        show_lidar=False,
        show_lane_line_detector=False,
        show_side_detector=False,
    ),

    # ===== engine config =====
    force_destroy=True,
    horizon=2500,
    # use_lateral_reward=True
)


class MixWaymoPGEnv(WaymoEnv):
    raise DeprecationWarning("Navigation error exists in this env, fix it")

    @classmethod
    def default_config(cls):
        config = super(MixWaymoPGEnv, cls).default_config()
        MIX_WAYMO_PG_ENV_CONFIG.update(
            dict(
                map_config={
                    BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
                    BaseMap.GENERATE_CONFIG: MIX_WAYMO_PG_ENV_CONFIG["block_num"],
                    # it can be a file path / block num / block ID sequence
                    BaseMap.LANE_WIDTH: 3.5,
                    BaseMap.LANE_NUM: 3,
                    "exit_length": 50,
                }
            )
        )
        config.update(MIX_WAYMO_PG_ENV_CONFIG)
        return config

    def __init__(self, config=None):
        super(MixWaymoPGEnv, self).__init__(config)
        self.waymo_map_manager = None
        self.waymo_traffic_manager = None
        self.pg_map_manager = None
        self.pg_traffic_manager = None

        self.total_environment = self.config["num_scenarios"] + self.config["num_scenarios"]
        self.real_data_ratio = self.config["num_scenarios"] / self.total_environment
        self.is_current_real_data = True

    def setup_engine(self):
        # Initialize all managers
        self.waymo_map_manager = WaymoMapManager()
        self.waymo_traffic_manager = WaymoTrafficManager()

        self.pg_map_manager = PGMapManager()
        self.pg_traffic_manager = PGTrafficManager()

        super(WaymoEnv, self).setup_engine()
        if self.real_data_ratio > 0:
            self.is_current_real_data = True
            self.engine.register_manager("data_manager", WaymoDataManager())
            self.engine.register_manager("map_manager", self.waymo_map_manager)
            if not self.config["no_traffic"]:
                self.engine.register_manager("traffic_manager", self.waymo_traffic_manager)
        else:
            self.is_current_real_data = False
            self.engine.register_manager("traffic_manager", self.pg_traffic_manager)
            self.engine.register_manager("map_manager", self.pg_map_manager)
            self._init_pg_episode()
        self.engine.accept("s", self.stop)
        self.engine.accept("q", self.switch_to_third_person_view)
        self.engine.accept("b", self.switch_to_top_down_view)

    def change_suite(self):
        if engine_initialized():
            # must lazy initialize at first
            if get_np_random(None).rand() < self.real_data_ratio:
                # change to real environment
                self.engine.update_manager("map_manager", self.waymo_map_manager, destroy_previous_manager=False)
                self.engine.update_manager(
                    "traffic_manager", self.waymo_traffic_manager, destroy_previous_manager=False
                )
                self.is_current_real_data = True
            else:
                self.is_current_real_data = False
                # change to PG environment
                self.engine.update_manager("map_manager", self.pg_map_manager, destroy_previous_manager=False)
                self.engine.update_manager("traffic_manager", self.pg_traffic_manager, destroy_previous_manager=False)
                self._init_pg_episode()

    def _init_pg_episode(self):
        self.config["target_vehicle_configs"]["default_agent"]["spawn_lane_index"] = (
            FirstPGBlock.NODE_1, FirstPGBlock.NODE_2, self.engine.np_random.randint(3)
        )
        self.config["target_vehicle_configs"]["default_agent"]["destination"] = None

    def reset(self, seed: Union[None, int] = None):
        self.change_suite()
        # ===== same as BaseEnv =====
        self.lazy_init()  # it only works the first time when reset() is called to avoid the error when render
        self._reset_global_seed(seed)
        if self.engine is None:
            raise ValueError(
                "Current MetaDrive instance is broken. Please make sure there is only one active MetaDrive "
                "environment exists in one process. You can try to call env.close() and then call "
                "env.reset() to rescue this environment. However, a better and safer solution is to check the "
                "singleton of MetaDrive and restart your program."
            )
        self.engine.reset()
        if self._top_down_renderer is not None:
            self._top_down_renderer.reset(self.current_map)

        self.dones = {agent_id: False for agent_id in self.vehicles.keys()}
        self.episode_rewards = defaultdict(float)
        self.episode_lengths = defaultdict(int)

        assert (len(self.vehicles) == self.num_agents) or (self.num_agents == -1)
        # ^^^^^^ same as Base Env ^^^^^

        if not self.is_current_real_data:
            # give a initial speed when on metadrive
            self.vehicle.set_velocity(self.vehicle.heading, self.engine.np_random.randint(10))
        return self._get_reset_return()

    def _reset_global_seed(self, force_seed=None):
        current_seed = force_seed if force_seed is not None else get_np_random(None).randint(
            self.config["start_scenario_index"], self.config["start_scenario_index"] +
            self.config["num_scenarios"] if self.is_current_real_data else self.config["num_scenarios"]
        )
        self.seed(current_seed)

    def done_function(self, vehicle_id: str):
        if self.is_current_real_data:
            return super(MixWaymoPGEnv, self).done_function(vehicle_id)
        else:
            return MetaDriveEnv.done_function(self, vehicle_id)

    def reward_function(self, vehicle_id: str):
        if self.is_current_real_data:
            return super(MixWaymoPGEnv, self).reward_function(vehicle_id)
        else:
            return MetaDriveEnv.reward_function(self, vehicle_id)

    def _is_out_of_road(self, vehicle):
        if self.is_current_real_data:
            return super(MixWaymoPGEnv, self)._is_out_of_road(vehicle)
        else:
            return vehicle.on_yellow_continuous_line or vehicle.crash_sidewalk or (not vehicle.on_lane)


class MixWaymoPGEnvWrapper(MixWaymoPGEnv):
    """
    This class is a convinient interface receive
    {"real_data_ratio": xyz,
    "total_num_scenarios":xyz} as input
    """
    TOTAL_SCENARIO = 100  # default 100

    def __init__(self, config=None):
        assert "data_directory" in config, "tell me waymo data path please"
        assert "real_data_ratio" in config, "tell me waymo data path please"
        env_config = config.copy()
        ratio = config["real_data_ratio"]
        assert 0 <= ratio <= 1, "ratio should be in [0, 1]"
        env_config["num_scenarios"] = int(config.get("total_num_scenarios", self.TOTAL_SCENARIO) * ratio)
        env_config["num_scenarios"] = int(
            config.get("total_num_scenarios", self.TOTAL_SCENARIO) - env_config["num_scenarios"]
        )
        if "real_data_ratio" in env_config:
            env_config.pop("real_data_ratio")
        if "total_num_scenarios" in env_config:
            env_config.pop("total_num_scenarios")
        super(MixWaymoPGEnvWrapper, self).__init__(env_config)


if __name__ == "__main__":
    env = MixWaymoPGEnvWrapper(
        dict(
            manual_control=True,
            use_render=True,
            data_directory=AssetLoader.file_path("waymo", return_raw_style=False),
            # # start_scenario=32,
            total_num_scenarios=10,
            real_data_ratio=0.3
        )
    )
    env.reset()
    while True:
        o, r, tm, tc, i = env.step(env.action_space.sample())
        env.render(text={"ts": env.episode_step})
        if tm or tc:
            env.reset()



import logging
from typing import Tuple

from panda3d.bullet import BulletBoxShape
from panda3d.bullet import BulletCylinderShape

from metadrive.component.static_object.base_static_object import BaseStaticObject
from metadrive.constants import CollisionGroup
from metadrive.constants import MetaDriveType
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.physics_node import BaseRigidBodyNode

LaneIndex = Tuple[str, str, int]

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class TrafficObject(BaseStaticObject):
    """
    Common interface for objects that appear on the road, beside vehicles.
    """
    CLASS_NAME = MetaDriveType.TRAFFIC_OBJECT
    COLLISION_MASK = CollisionGroup.TrafficObject

    COST_ONCE = True  # cost will give at the first time

    def __init__(self, position, heading_theta, lane=None, random_seed=None, name=None):
        """
        :param lane: the lane to spawn object
        """
        assert self.CLASS_NAME is not None, "Assign a name for this class for finding it easily"
        super(TrafficObject, self).__init__(position, heading_theta, lane, random_seed, name=name)
        self.crashed = False

    def reset(self, position, heading_theta, lane=None, random_seed=None, name=None, *args, **kwargs):
        self.crashed = False
        super(TrafficObject, self).reset(position, heading_theta, lane, random_seed, name, *args, **kwargs)


class TrafficCone(TrafficObject):
    """Placed near the construction section to indicate that traffic is prohibited"""

    RADIUS = 0.2
    HEIGHT = 2
    MASS = 2
    CLASS_NAME = MetaDriveType.TRAFFIC_CONE

    def __init__(self, position, heading_theta, lane=None, static: bool = False, random_seed=None, name=None):
        super(TrafficCone, self).__init__(position, heading_theta, lane, random_seed, name)

        n = BaseRigidBodyNode(self.name, self.CLASS_NAME)
        self.add_body(n)

        self.body.addShape(BulletCylinderShape(self.RADIUS, self.HEIGHT))
        self.set_static(static)
        if self.render:
            model = self.loader.loadModel(AssetLoader.file_path("models", "traffic_cone", "scene.gltf"))
            model.setScale(0.02, 0.02, 0.025)
            model.setPos(0, 0, -self.HEIGHT / 2 + 0.05)
            model.reparentTo(self.origin)

    @property
    def top_down_length(self):
        return self.RADIUS * 4

    @property
    def top_down_width(self):
        return self.RADIUS * 4

    @property
    def top_down_color(self):
        return 235, 84, 42

    @property
    def LENGTH(self):
        return self.RADIUS

    @property
    def WIDTH(self):
        return self.RADIUS


class TrafficWarning(TrafficObject):
    """Placed behind the vehicle when it breaks down"""

    HEIGHT = 1.2
    MASS = 1
    RADIUS = 0.5

    def __init__(self, position, heading_theta, lane=None, static: bool = False, random_seed=None, name=None):
        super(TrafficWarning, self).__init__(position, heading_theta, lane, random_seed, name)

        n = BaseRigidBodyNode(self.name, self.CLASS_NAME)
        self.add_body(n)

        self.body.addShape(BulletCylinderShape(self.RADIUS, self.HEIGHT))
        self.set_static(static)
        if self.render:
            model = self.loader.loadModel(AssetLoader.file_path("models", "warning", "warning.gltf"))
            model.setScale(0.02)
            model.setH(-90)
            model.setPos(0, 0, -self.HEIGHT / 2 - 0.1)
            model.reparentTo(self.origin)

    @property
    def top_down_length(self):
        return self.RADIUS * 2

    @property
    def top_down_width(self):
        return self.RADIUS * 2

    @property
    def LENGTH(self):
        return self.RADIUS

    @property
    def WIDTH(self):
        return self.RADIUS


class TrafficBarrier(TrafficObject):
    """A barrier"""

    HEIGHT = 2.0
    MASS = 10
    CLASS_NAME = MetaDriveType.TRAFFIC_BARRIER

    def __init__(self, position, heading_theta, lane=None, static: bool = False, random_seed=None, name=None):
        super(TrafficBarrier, self).__init__(position, heading_theta, lane, random_seed, name)
        n = BaseRigidBodyNode(self.name, self.CLASS_NAME)
        self.add_body(n)

        self.body.addShape(BulletBoxShape((self.WIDTH / 2, self.LENGTH / 2, self.height / 2)))
        self.set_static(static)
        if self.render:
            model = self.loader.loadModel(AssetLoader.file_path("models", "barrier", "scene.gltf"))
            model.setH(-90)
            model.setPos(0, 0, -0.93)
            model.setScale(0.7)
            model.reparentTo(self.origin)

    @property
    def LENGTH(self):
        return 2.0

    @property
    def WIDTH(self):
        return 0.3

    @property
    def width(self):
        logger.warning("This API will be deprecated, Please use {}.WIDTH instead".format(self.class_name))
        return self.WIDTH

    @property
    def length(self):
        logger.warning("This API will be deprecated, Please use {}.LENGTH instead".format(self.class_name))
        return self.LENGTH

    @property
    def height(self):
        return self.HEIGHT

    @property
    def top_down_length(self):
        # reverse the direction
        return self.WIDTH * 2

    @property
    def top_down_width(self):
        # reverse the direction
        return self.LENGTH



import copy
import logging
import uuid
from metadrive.constants import CamMask
import gymnasium as gym
import numpy as np
from panda3d.core import NodePath, Material, LVector4
from metadrive.base_class.configurable import Configurable
from metadrive.base_class.randomizable import Randomizable
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import get_engine


class BasePolicy(Randomizable, Configurable):
    DEBUG_MARK_COLOR = (255, 255, 255, 255)
    DEBUG_MARK_MODEL = None
    SYNC_DEBUG_MARK_POS_TASK_NAME = "policy_mark"

    def __init__(self, control_object, random_seed=None, config=None):
        Randomizable.__init__(self, random_seed)
        Configurable.__init__(self, config)
        # self.engine = get_engine()
        self.control_object = control_object
        self.action_info = dict()
        self._debug_mark = None
        self._mark_update_task_name = None
        self.show_policy_mark()

    def _get_task_name(self):
        return self.SYNC_DEBUG_MARK_POS_TASK_NAME + str(uuid.uuid4())

    def act(self, *args, **kwargs):
        """
        Return action [], policy implement information (dict) can be written in self.action_info, which will be
        retrieved automatically
        """
        pass

    def _sync_debug_mark(self, task):
        assert self._debug_mark is not None
        assert self._mark_update_task_name is not None
        pos = self.control_object.origin.getPos()
        height = getattr(self.control_object, "HEIGHT") if hasattr(self.control_object, "HEIGHT") else 2
        self._debug_mark.setPos(pos[0], pos[1], height + 0.5)
        return task.cont

    def get_action_info(self):
        """
        Get current action info for env.step() retrieve
        """
        return copy.deepcopy(self.action_info)

    def reset(self):
        self.action_info.clear()

    def destroy(self):
        if self._debug_mark is not None:
            self.engine.taskMgr.remove(self._mark_update_task_name)
            self._debug_mark.removeNode()
            self._debug_mark = None
            self._mark_update_task_name = None
        super(BasePolicy, self).destroy()
        self.control_object = None
        logging.debug("{} is released".format(self.__class__.__name__))

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.name

    @property
    def engine(self):
        return get_engine()

    @classmethod
    def get_input_space(cls):
        """
        It defines the input space of this class of policy
        """
        # logging.info(
        #     "No input space set for this policy! If you are querying an action space, "
        #     "the agent policy may not take any external input from env.step() and thus the env.action_space is None"
        # )
        return gym.spaces.Box(-1.0, 1.0, shape=(2, ), dtype=np.float32)

    def get_state(self):
        return self.get_action_info()

    @property
    def episode_step(self):
        return self.engine.episode_step

    def show_policy_mark(self):
        if not self.engine.global_config["show_policy_mark"]:
            return
        if self.DEBUG_MARK_MODEL is None:
            self.DEBUG_MARK_MODEL = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
            self.DEBUG_MARK_MODEL.setScale(0.5)
            self.DEBUG_MARK_MODEL.show(CamMask.MainCam)
            self.DEBUG_MARK_MODEL.hprInterval(5, (360, 0, 0)).loop()
        self._debug_mark = NodePath(self.name)
        self.DEBUG_MARK_MODEL.instanceTo(self._debug_mark)

        # texture = AssetLoader.loader.loadTexture(AssetLoader.file_path("textures", "height_map.png"))
        # texture.set_format(Texture.F_srgb)
        # self._debug_mark.setTexture(texture)
        r, g, b, a = self.DEBUG_MARK_COLOR
        material = Material()
        material.setBaseColor(LVector4(r / 255, g / 255, b / 255, a / 255))

        self._debug_mark.setMaterial(material)

        self._debug_mark.reparentTo(self.engine.origin)
        pos = self.control_object.origin.getPos()
        height = getattr(self.control_object, "HEIGHT") if hasattr(self.control_object, "HEIGHT") else 2
        self._debug_mark.setPos(pos[0], pos[1], height + 0.5)
        self._mark_update_task_name = self._get_task_name()
        self.engine.taskMgr.add(self._sync_debug_mark, self._mark_update_task_name)



from metadrive.utils.config import Config, merge_config_with_unknown_keys, merge_config
from metadrive.utils.coordinates_shift import panda_heading, panda_vector, metadrive_heading, metadrive_vector
from metadrive.utils.math import safe_clip, clip, norm, distance_greater, safe_clip_for_small_array, Vector
from metadrive.utils.random import get_np_random, random_string
from metadrive.utils.registry import get_metadrive_class
from metadrive.utils.utils import is_mac, import_pygame, recursive_equal, setup_logger, merge_dicts, \
    concat_step_infos, is_win



import logging
import math
import time

import numpy as np
from panda3d.core import NodePath, TextNode, LQuaternionf

from metadrive.constants import COLLISION_INFO_COLOR, COLOR, MetaDriveType, \
    CamMask, RENDER_MODE_NONE
from metadrive.engine.asset_loader import AssetLoader


class DisplayRegionPosition:
    left = [0., 1 / 3, 0.8, 1.0]
    mid = [1 / 3, 2 / 3, 0.8, 1.0]
    right = [2 / 3, 1, 0.8, 1.0]


class Interface:
    """
    Visualization interface, state banner and vehicle panel
    """
    ARROW_COLOR = COLLISION_INFO_COLOR["green"][1]

    def __init__(self, base_engine):
        self._node_path_list = []
        # self.engine = base_engine
        self.vehicle_panel = None
        self.right_panel = None
        self.mid_panel = None
        self.left_panel = None
        self.contact_result_render = None
        self.arrow = None
        self._left_arrow = None
        self._right_arrow = None
        self._contact_banners = {}  # to save time/memory
        self.current_banner = None
        self.need_interface = base_engine.mode != RENDER_MODE_NONE and not base_engine.global_config[
            "debug_physics_world"] and base_engine.global_config["show_interface"]
        if base_engine.mode == RENDER_MODE_NONE:
            assert self.need_interface is False, \
                "We should not using interface with extra cameras when in offscreen mode!"
        self.init_interface()
        self._is_showing_arrow = True  # store the state of navigation mark

    def after_step(self):
        if self.engine.current_track_vehicle is not None and self.need_interface and self.engine.mode != RENDER_MODE_NONE:
            track_v = self.engine.current_track_vehicle
            if self.vehicle_panel is not None:
                self.vehicle_panel.update_vehicle_state(track_v)
            self._render_contact_result(track_v.contact_results)
            if hasattr(track_v, "navigation") and track_v.navigation is not None:
                self._update_navi_arrow(track_v.navigation.navi_arrow_dir)

    def init_interface(self):
        if self.need_interface:
            info_np = NodePath("Collision info nodepath")
            info_np.reparentTo(self.engine.aspect2d)

            self._node_path_list.append(info_np)

            self.contact_result_render = info_np
            for idx, panel_name, in enumerate(reversed(self.engine.global_config["interface_panel"])):
                if idx == 0:
                    self.right_panel = self.engine.get_sensor(panel_name)
                elif idx == 1:
                    self.mid_panel = self.engine.get_sensor(panel_name)
                elif idx == 2:
                    self.left_panel = self.engine.get_sensor(panel_name)
                else:
                    raise ValueError("Can not add > 3 panels!")
                if panel_name == "dashboard":
                    self.vehicle_panel = self.engine.get_sensor(panel_name)

            self.arrow = self.engine.aspect2d.attachNewNode("arrow")
            self._node_path_list.append(self.arrow)

            navi_arrow_model = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "navi_arrow.gltf"))
            navi_arrow_model.setScale(0.1, 0.12, 0.2)
            navi_arrow_model.setPos(2, 1.15, -0.221)
            self._left_arrow = self.arrow.attachNewNode("left arrow")
            self._node_path_list.append(self._left_arrow)

            self._left_arrow.setP(180)
            self._right_arrow = self.arrow.attachNewNode("right arrow")
            self._node_path_list.append(self._right_arrow)

            self._left_arrow.setColor(self.ARROW_COLOR)
            self._right_arrow.setColor(self.ARROW_COLOR)
            if self.engine.global_config["show_interface_navi_mark"]:
                navi_arrow_model.instanceTo(self._left_arrow)
                navi_arrow_model.instanceTo(self._right_arrow)
            self._left_arrow.detachNode()
            self._right_arrow.detachNode()
            self.arrow.setPos(0, 0, 0.08)
            self.arrow.hide(CamMask.AllOn)
            self.arrow.show(CamMask.MainCam)
            self.arrow.setQuat(LQuaternionf(math.cos(-math.pi / 4), 0, 0, math.sin(-math.pi / 4)))
            # the transparency attribute of gltf model is invalid on windows
            # self.arrow.setTransparency(TransparencyAttrib.M_alpha)

    def stop_track(self):
        if self.need_interface:
            if self.right_panel:
                self.right_panel.remove_display_region()
            if self.mid_panel:
                self.mid_panel.remove_display_region()
            if self.left_panel:
                self.left_panel.remove_display_region()
            self.contact_result_render.detachNode()
            self.arrow.detachNode()

    def track(self, vehicle):
        if self.need_interface:
            if self.right_panel:
                self.right_panel.add_display_region(DisplayRegionPosition.right)
                self.right_panel.track(vehicle)
            if self.mid_panel:
                self.mid_panel.add_display_region(DisplayRegionPosition.mid)
                self.mid_panel.track(vehicle)
            if self.left_panel:
                self.left_panel.add_display_region(DisplayRegionPosition.left)
                self.left_panel.track(vehicle)

            self.arrow.reparentTo(self.engine.aspect2d)
            self.contact_result_render.reparentTo(self.engine.aspect2d)

    def _render_banner(self, text, color=COLLISION_INFO_COLOR["green"][1]):
        """
        Render the banner in the left bottom corner.
        """
        if self.contact_result_render is None:
            return
        if self.current_banner is not None:
            self.current_banner.detachNode()
        if text in self._contact_banners:
            self._contact_banners[text].reparentTo(self.contact_result_render)
            self.current_banner = self._contact_banners[text]
        else:
            new_banner = NodePath(TextNode("collision_info:{}".format(text)))

            self._node_path_list.append(new_banner)

            self._contact_banners[text] = new_banner
            text_node = new_banner.node()
            text_node.setCardColor(color)
            text_node.setText(text)
            text_node.setTextScale(0.96)
            text_node.setCardActual(-9, 9.1, -0.26, 1)
            text_node.setCardDecal(True)
            text_node.setTextColor(1, 1, 1, 1)
            text_node.setAlign(TextNode.A_center)
            new_banner.setScale(0.05 * 3 / 4 * self.engine.w_scale)
            new_banner.setPos(-0.662 * self.engine.w_scale, 0, -0.987 * self.engine.h_scale)
            # new_banner.setPos(-0.75 * self.engine.w_scale, 0, -0.8 * self.engine.h_scale)
            new_banner.reparentTo(self.contact_result_render)
            self.current_banner = new_banner

    def _render_contact_result(self, contacts):
        contacts = sorted(list(contacts), key=lambda c: COLLISION_INFO_COLOR[COLOR[c]][0])
        text = contacts[0] if len(contacts) != 0 else None
        if text is None:
            text = MetaDriveType.LANE_SURFACE_STREET \
                if time.time() - self.engine._episode_start_time > 10 else "Press H to see help message"
            self._render_banner(text, COLLISION_INFO_COLOR["green"][1])
        else:
            if text == MetaDriveType.VEHICLE:
                text = MetaDriveType.VEHICLE
            self._render_banner(text, COLLISION_INFO_COLOR[COLOR[text]][1])

    def destroy(self):

        for np in self._node_path_list:
            np.detachNode()
            np.removeNode()

        if self.need_interface:
            self.stop_track()
            self.contact_result_render.removeNode()
            self.arrow.removeNode()
            self.contact_result_render = None
            self._contact_banners = None
            self.current_banner = None
            if self.right_panel:
                self.right_panel.destroy()
            if self.mid_panel:
                self.mid_panel.destroy()
            if self.left_panel:
                self.left_panel.destroy()

    def _update_navi_arrow(self, lanes_heading):
        lane_0_heading = lanes_heading[0]
        lane_1_heading = lanes_heading[1]
        if abs(lane_0_heading - lane_1_heading) < 0.01:
            if self._is_showing_arrow:
                self._left_arrow.detachNode()
                self._right_arrow.detachNode()
                self._is_showing_arrow = False
        else:
            dir_0 = np.array([math.cos(lane_0_heading), math.sin(lane_0_heading), 0])
            dir_1 = np.array([math.cos(lane_1_heading), math.sin(lane_1_heading), 0])
            cross_product = np.cross(dir_1, dir_0)
            left = False if cross_product[-1] < 0 else True
            if not self._is_showing_arrow:
                self._is_showing_arrow = True
            if left:
                if not self._left_arrow.hasParent():
                    self._left_arrow.reparentTo(self.arrow)
                if self._right_arrow.hasParent():
                    self._right_arrow.detachNode()
            else:
                if not self._right_arrow.hasParent():
                    self._right_arrow.reparentTo(self.arrow)
                if self._left_arrow.hasParent():
                    self._left_arrow.detachNode()

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import division

from panda3d.core import Vec3, Camera, OrthographicLens, PTAMat4
from panda3d.core import LVecBase2i

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.render_stage import RenderStage
from metadrive.render_pipeline.rpcore.util.generic import snap_shadow_map


class PSSMDistShadowStage(RenderStage):
    """ This stage generates a depth map using Variance Shadow Maps for very
    distant objects. """

    required_inputs = []

    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self.resolution = 2048
        self.clip_size = 500
        self.sun_distance = 8000
        self.sun_vector = Vec3(0, 0, 1)
        self.pta_mvp = PTAMat4.empty_array(1)

    @property
    def produced_inputs(self):
        return {"PSSMDistSunShadowMapMVP": self.pta_mvp}

    @property
    def produced_pipes(self):
        return {"PSSMDistSunShadowMap": self.target_blur_h.color_tex}

    @property
    def mvp(self):
        return Globals.base.render.get_transform(self.cam_node).get_mat() * \
            self.cam_lens.get_projection_mat()

    def update(self):
        self.target.active = False
        self.target_convert.active = False
        self.target_blur_v.active = False
        self.target_blur_h.active = False

        # Query scheduled tasks
        if self._pipeline.task_scheduler.is_scheduled("pssm_distant_shadows"):

            self.target.active = True

            # Reposition camera before we capture the scene
            cam_pos = Globals.base.cam.get_pos(Globals.base.render)
            self.cam_node.set_pos(cam_pos + self.sun_vector * self.sun_distance)
            self.cam_node.look_at(cam_pos)
            self.cam_lens.set_film_size(self.clip_size, self.clip_size)

            snap_shadow_map(self.mvp, self.cam_node, self.resolution)

        if self._pipeline.task_scheduler.is_scheduled("pssm_convert_distant_to_esm"):
            self.target_convert.active = True
        if self._pipeline.task_scheduler.is_scheduled("pssm_blur_distant_vert"):
            self.target_blur_v.active = True
        if self._pipeline.task_scheduler.is_scheduled("pssm_blur_distant_horiz"):
            self.target_blur_h.active = True

            # Only update the MVP as soon as the shadow map is available
            self.pta_mvp[0] = self.mvp

    def create(self):
        self.camera = Camera("PSSMDistShadowsESM")
        self.cam_lens = OrthographicLens()
        self.cam_lens.set_film_size(12000, 12000)
        self.cam_lens.set_near_far(10.0, self.sun_distance * 2)
        self.camera.set_lens(self.cam_lens)
        self.cam_node = Globals.base.render.attach_new_node(self.camera)

        self.target = self.create_target("ShadowMap")
        self.target.size = self.resolution
        self.target.add_depth_attachment(bits=32)
        self.target.prepare_render(self.cam_node)

        self.target_convert = self.create_target("ConvertToESM")
        self.target_convert.size = self.resolution
        self.target_convert.add_color_attachment(bits=(32, 0, 0, 0))
        self.target_convert.prepare_buffer()
        self.target_convert.set_shader_input("SourceTex", self.target.depth_tex)

        self.target_blur_v = self.create_target("BlurVert")
        self.target_blur_v.size = self.resolution
        self.target_blur_v.add_color_attachment(bits=(32, 0, 0, 0))
        self.target_blur_v.prepare_buffer()
        self.target_blur_v.set_shader_inputs(SourceTex=self.target_convert.color_tex, direction=LVecBase2i(1, 0))

        self.target_blur_h = self.create_target("BlurHoriz")
        self.target_blur_h.size = self.resolution
        self.target_blur_h.add_color_attachment(bits=(32, 0, 0, 0))
        self.target_blur_h.prepare_buffer()
        self.target_blur_h.set_shader_inputs(SourceTex=self.target_blur_v.color_tex, direction=LVecBase2i(0, 1))

        # Register shadow camera
        self._pipeline.tag_mgr.register_camera("shadow", self.camera)

    def reload_shaders(self):
        self.target_convert.shader = self.load_plugin_shader("convert_to_esm.frag.glsl")
        self.target_blur_v.shader = self.load_plugin_shader("blur_esm.frag.glsl")
        self.target_blur_h.shader = self.load_plugin_shader("blur_esm.frag.glsl")

    def set_shader_input(self, *args):
        Globals.render.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        Globals.render.set_shader_inputs(**kwargs)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rpcore.render_stage import RenderStage


class FinalStage(RenderStage):
    """ This stage is the final stage and outputs the shaded scene to the
    screen """

    required_pipes = ["ShadedScene"]

    @property
    def produced_pipes(self):
        return {"ShadedScene": self.target.color_tex}

    def create(self):

        self.target = self.create_target("FinalStage")
        self.target.add_color_attachment(bits=16)
        self.target.prepare_buffer()

        # XXX: We cannot simply assign the final shader to the window display
        # region. This is because of a bug that the last FBO always only gets
        # 8 bits, regardles of what was requested - probably because of the
        # assumption that no tonemapping/srgb correction will follow afterwards.
        #
        # This might be a driver bug, or an optimization in Panda3D. However, it
        # also has the nice side effect that when taking screenshots (or using
        # the pixel inspector), we get the srgb corrected data, so its not too
        # much of a disadvantage.
        self.present_target = self.create_target("FinalPresentStage")
        self.present_target.present_on_screen()
        self.present_target.set_shader_input("SourceTex", self.target.color_tex)

    def reload_shaders(self):
        self.target.shader = self.load_shader("final_stage.frag.glsl")
        self.present_target.shader = self.load_shader("final_present_stage.frag.glsl")



"""

Wrapper script to import all qt classes

"""

from __future__ import print_function
import sys

PYQT_VERSION = 5

try:
    import PyQt5
except ImportError:
    print("Could not import PyQt5, trying to import PyQt4")
    try:
        import PyQt4
        PYQT_VERSION = 4
    except ImportError:
        print("Failed to import either PyQt4 or PyQt5!")
        sys.exit(-1)

if PYQT_VERSION == 4:
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *

    def qt_connect(obj, signal_name, handler):
        QObject.connect(obj, SIGNAL(signal_name), handler)

elif PYQT_VERSION == 5:
    from PyQt5.QtGui import *
    from PyQt5.QtWidgets import *
    from PyQt5.QtCore import *

    def qt_connect(obj, signal_name, handler):
        if "(" in signal_name:
            slot_name = signal_name[:signal_name.index("(")]
            handler_type = signal_name[signal_name.index("(") + 1:].rstrip(")")
            handler_type = str(handler_type.replace("*", "")).strip()
            if not handler_type:
                getattr(obj, slot_name).connect(handler)
            else:
                py_type = {
                    "int": int,
                    "QString": str,
                    "double": float,
                    "QListWidgetItem": QListWidgetItem
                }[handler_type]
                getattr(obj, slot_name)[py_type].connect(handler)
        else:
            getattr(obj, signal_name).connect(handler)


def qt_register_fonts():
    import os
    this_dir = os.path.realpath(os.path.dirname(__file__))
    font_dir = os.path.join(this_dir, "..", "data", "font")
    for f in os.listdir(font_dir):
        if f.endswith(".ttf"):
            fpath = os.path.relpath(os.path.join(font_dir, f))
            QFontDatabase.addApplicationFont(fpath)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.yaml import load_yaml_file
from metadrive.render_pipeline.rpcore.rpobject import RPObject


class TaskScheduler(RPObject):
    """ This class manages the scheduled tasks and splits them over multiple
    frames. Plugins can query whether their subtasks should be executed
    or queued for later frames. Also performs analysis on the task configuration
    to figure if tasks are distributed uniformly. """
    def __init__(self, pipeline):
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._tasks = []
        self._frame_index = 0
        self._load_config()

    def _load_config(self):
        """ Loads the tasks distribution configuration """
        config = load_yaml_file("/$$rpconfig/task-scheduler.yaml")["frame_cycles"]
        for frame_name, tasks in config:  # pylint: disable=unused-variable
            self._tasks.append(tasks)

    def _check_missing_schedule(self, task_name):
        """ Checks whether the given task is scheduled at some point. This can
        be used to check whether any task is missing in the task scheduler config. """
        for tasks in self._tasks:
            if task_name in tasks:
                break
        else:
            self.error("Task '" + task_name + "' is never scheduled and thus will never run!")

    def is_scheduled(self, task_name):
        """ Returns whether a given task is supposed to run this frame """
        self._check_missing_schedule(task_name)
        return task_name in self._tasks[self._frame_index]

    def step(self):
        """ Advances one frame """
        self._frame_index = (self._frame_index + 1) % len(self._tasks)

    @property
    def num_tasks(self):
        """ Returns the total amount of tasks """
        return sum((len(i) for i in self._tasks))

    @property
    def num_scheduled_tasks(self):
        """ Returns the amount of scheduled tasks this frame """
        return len(self._tasks[self._frame_index])



import random

import matplotlib.pyplot as plt

from metadrive import MetaDriveEnv
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.utils.draw_top_down_map import draw_top_down_map

if __name__ == '__main__':
    env = MetaDriveEnv(config=dict(num_scenarios=100, map=7, start_seed=0))
    fig, axs = plt.subplots(2, 3, figsize=(10, 10), dpi=100)
    count = 0
    print("We are going to draw 6 maps! 3 for PG maps, 3 for real world ones!")
    for i in range(2):
        if i == 1:
            env.close()
            env = ScenarioEnv(dict(start_scenario_index=0, num_scenarios=3))
        for j in range(3):
            count += 1
            env.reset(seed=j)
            m = draw_top_down_map(env.current_map)
            # m = env.get_map()
            ax = axs[i][j]
            ax.imshow(m, cmap="bone")
            ax.set_xticks([])
            ax.set_yticks([])
            print("Drawing {}-th map!".format(count))
    fig.suptitle("Top-down view of generated maps")
    plt.show()
    env.close()



TrafficSignal = {
    0: 'LANE_STATE_UNKNOWN',
    1: 'LANE_STATE_ARROW_STOP',
    2: 'LANE_STATE_ARROW_CAUTION',
    3: 'LANE_STATE_ARROW_GO',
    4: 'LANE_STATE_STOP',
    5: 'LANE_STATE_CAUTION',
    6: 'LANE_STATE_GO',
    7: 'LANE_STATE_FLASHING_STOP',
    8: 'LANE_STATE_FLASHING_CAUTION'
}


class WaymoLaneType:
    UNKNOWN = 0
    LANE_FREEWAY = 1
    LANE_SURFACE_STREET = 2
    LANE_BIKE_LANE = 3

    ENUM_TO_STR = {
        UNKNOWN: 'UNKNOWN',
        LANE_FREEWAY: 'LANE_FREEWAY',
        LANE_SURFACE_STREET: 'LANE_SURFACE_STREET',
        LANE_BIKE_LANE: 'LANE_BIKE_LANE'
    }

    @classmethod
    def from_waymo(cls, item):
        return cls.ENUM_TO_STR[item]


class WaymoRoadLineType:
    UNKNOWN = 0
    BROKEN_SINGLE_WHITE = 1
    SOLID_SINGLE_WHITE = 2
    SOLID_DOUBLE_WHITE = 3
    BROKEN_SINGLE_YELLOW = 4
    BROKEN_DOUBLE_YELLOW = 5
    SOLID_SINGLE_YELLOW = 6
    SOLID_DOUBLE_YELLOW = 7
    PASSING_DOUBLE_YELLOW = 8

    ENUM_TO_STR = {
        UNKNOWN: 'UNKNOWN',
        BROKEN_SINGLE_WHITE: 'ROAD_LINE_BROKEN_SINGLE_WHITE',
        SOLID_SINGLE_WHITE: 'ROAD_LINE_SOLID_SINGLE_WHITE',
        SOLID_DOUBLE_WHITE: 'ROAD_LINE_SOLID_DOUBLE_WHITE',
        BROKEN_SINGLE_YELLOW: 'ROAD_LINE_BROKEN_SINGLE_YELLOW',
        BROKEN_DOUBLE_YELLOW: 'ROAD_LINE_BROKEN_DOUBLE_YELLOW',
        SOLID_SINGLE_YELLOW: 'ROAD_LINE_SOLID_SINGLE_YELLOW',
        SOLID_DOUBLE_YELLOW: 'ROAD_LINE_SOLID_DOUBLE_YELLOW',
        PASSING_DOUBLE_YELLOW: 'ROAD_LINE_PASSING_DOUBLE_YELLOW'
    }

    @classmethod
    def from_waymo(cls, item):
        return cls.ENUM_TO_STR[item]


class WaymoRoadEdgeType:
    UNKNOWN = 0
    # Physical road boundary that doesn't have traffic on the other side (e.g., a curb or the k-rail on the right side of a freeway).
    BOUNDARY = 1
    # Physical road boundary that separates the car from other traffic (e.g. a k-rail or an island).
    MEDIAN = 2

    ENUM_TO_STR = {UNKNOWN: 'UNKNOWN', BOUNDARY: 'ROAD_EDGE_BOUNDARY', MEDIAN: 'ROAD_EDGE_MEDIAN'}

    @classmethod
    def from_waymo(cls, item):
        return cls.ENUM_TO_STR[item]


class WaymoAgentType:
    UNSET = 0
    VEHICLE = 1
    PEDESTRIAN = 2
    CYCLIST = 3
    OTHER = 4

    ENUM_TO_STR = {UNSET: 'UNSET', VEHICLE: 'VEHICLE', PEDESTRIAN: 'PEDESTRIAN', CYCLIST: 'CYCLIST', OTHER: 'OTHER'}

    @classmethod
    def from_waymo(cls, item):
        return cls.ENUM_TO_STR[item]



class Token(object):
    def __init__(self, start_mark, end_mark):
        self.start_mark = start_mark
        self.end_mark = end_mark

    def __repr__(self):
        attributes = [key for key in self.__dict__ if not key.endswith('_mark')]
        attributes.sort()
        arguments = ', '.join(['%s=%r' % (key, getattr(self, key)) for key in attributes])
        return '%s(%s)' % (self.__class__.__name__, arguments)


#class BOMToken(Token):
#    id = '<byte order mark>'


class DirectiveToken(Token):
    id = '<directive>'

    def __init__(self, name, value, start_mark, end_mark):
        self.name = name
        self.value = value
        self.start_mark = start_mark
        self.end_mark = end_mark


class DocumentStartToken(Token):
    id = '<document start>'


class DocumentEndToken(Token):
    id = '<document end>'


class StreamStartToken(Token):
    id = '<stream start>'

    def __init__(self, start_mark=None, end_mark=None, encoding=None):
        self.start_mark = start_mark
        self.end_mark = end_mark
        self.encoding = encoding


class StreamEndToken(Token):
    id = '<stream end>'


class BlockSequenceStartToken(Token):
    id = '<block sequence start>'


class BlockMappingStartToken(Token):
    id = '<block mapping start>'


class BlockEndToken(Token):
    id = '<block end>'


class FlowSequenceStartToken(Token):
    id = '['


class FlowMappingStartToken(Token):
    id = '{'


class FlowSequenceEndToken(Token):
    id = ']'


class FlowMappingEndToken(Token):
    id = '}'


class KeyToken(Token):
    id = '?'


class ValueToken(Token):
    id = ':'


class BlockEntryToken(Token):
    id = '-'


class FlowEntryToken(Token):
    id = ','


class AliasToken(Token):
    id = '<alias>'

    def __init__(self, value, start_mark, end_mark):
        self.value = value
        self.start_mark = start_mark
        self.end_mark = end_mark


class AnchorToken(Token):
    id = '<anchor>'

    def __init__(self, value, start_mark, end_mark):
        self.value = value
        self.start_mark = start_mark
        self.end_mark = end_mark


class TagToken(Token):
    id = '<tag>'

    def __init__(self, value, start_mark, end_mark):
        self.value = value
        self.start_mark = start_mark
        self.end_mark = end_mark


class ScalarToken(Token):
    id = '<scalar>'

    def __init__(self, value, plain, start_mark, end_mark, style=None):
        self.value = value
        self.plain = plain
        self.start_mark = start_mark
        self.end_mark = end_mark
        self.style = style



import numpy as np

from metadrive.utils.cuda import check_cudart_err

_cuda_enable = True
try:
    import cupy as cp
    from OpenGL.GL import GL_TEXTURE_2D  # noqa F403
    from cuda import cudart
    from cuda.cudart import cudaGraphicsRegisterFlags
    from panda3d.core import GraphicsOutput, Texture, GraphicsStateGuardianBase, DisplayRegionDrawCallbackData
except ImportError:
    _cuda_enable = False
from panda3d.core import Vec3

from metadrive.engine.core.image_buffer import ImageBuffer


class BaseCamera(ImageBuffer):
    """
    To enable the image observation, set image_observation to True.
    Every objects share the same camera, to boost the efficiency and save memory.
    Camera configuration is read from the global config automatically.
    """
    # shape(dim_1, dim_2)
    BUFFER_W = 84  # dim 1
    BUFFER_H = 84  # dim 2
    CAM_MASK = None
    display_region_size = [1 / 3, 2 / 3, 0.8, 1.0]
    attached_object = None

    def __init__(self, engine, setup_pbr=False, need_cuda=False):
        self._enable_cuda = need_cuda
        super(BaseCamera, self).__init__(
            self.BUFFER_W, self.BUFFER_H, Vec3(0., 0.8, 1.5), self.BKG_COLOR, setup_pbr=setup_pbr, engine=engine
        )

        width = self.BUFFER_W
        height = self.BUFFER_H
        if (width > 100 or height > 100) and not self.enable_cuda:
            # Too large height or width will cause corruption in Mac.
            self.logger.warning(
                "You may using too large buffer! The height is {}, and width is {}. "
                "It may lower the sample efficiency! Considering reduce buffer size or using cuda image by"
                " set [image_on_cuda=True].".format(height, width)
            )
        self.cuda_graphics_resource = None
        if self.enable_cuda:
            assert _cuda_enable, "Can not enable cuda rendering pipeline"

            # returned tensor property
            self.cuda_dtype = np.uint8
            self.cuda_shape = (self.BUFFER_W, self.BUFFER_H)
            self.cuda_strides = None
            self.cuda_order = "C"

            self._cuda_buffer = None

            # make texture
            self.cuda_texture = Texture()
            self.buffer.addRenderTexture(self.cuda_texture, GraphicsOutput.RTMBindOrCopy)

            def _callback_func(cbdata: DisplayRegionDrawCallbackData):
                # print("DRAW CALLBACK!!!!!!!!!!!!!!!11")
                cbdata.upcall()
                if not self.registered and self.texture_context_future.done():
                    self.register()
                if self.registered:
                    with self as array:
                        self.cuda_rendered_result = array

            # Fill the buffer due to multi-thread
            for _ in range(3):
                self.engine.graphicsEngine.renderFrame()
            self.cam.node().getDisplayRegion(0).setDrawCallback(_callback_func)

            self.gsg = GraphicsStateGuardianBase.getDefaultGsg()
            self.texture_context_future = self.cuda_texture.prepare(self.gsg.prepared_objects)
            self.cuda_texture_identifier = None
            self.new_cuda_mem_ptr = None
            self.cuda_rendered_result = None

    @property
    def enable_cuda(self):
        return self is not None and self._enable_cuda

    def get_image(self, base_object):
        """
        Borrow the camera to get observations
        """
        self.origin.reparentTo(base_object.origin)
        ret = super(BaseCamera, self).get_image()
        self.track(self.attached_object)
        return ret

    def save_image(self, base_object, name="debug.png"):
        img = self.get_image(base_object)
        img.write(name)

    def get_pixels_array(self, base_object, clip=True) -> np.ndarray:
        self.track(base_object)
        if self.enable_cuda:
            assert self.cuda_rendered_result is not None
            ret = self.cuda_rendered_result[..., :-1][..., ::-1][::-1]
        else:
            ret = self.get_rgb_array_cpu()
        if self.engine.global_config["rgb_to_grayscale"]:
            ret = np.dot(ret[..., :3], [0.299, 0.587, 0.114])
        if not clip:
            return ret.astype(np.uint8)
        else:
            return ret / 255

    def destroy(self):
        if self.registered:
            self.unregister()
        ImageBuffer.destroy(self)

    def get_cam(self):
        return self.cam

    def get_lens(self):
        return self.lens

    # # following functions are for onscreen render
    # def add_display_region(self, display_region):
    #     super(BaseCamera, self).add_display_region(display_region)

    def remove_display_region(self):
        super(BaseCamera, self).remove_display_region()

    def track(self, base_object):
        if base_object is not None and self is not None:
            self.attached_object = base_object
            self.origin.reparentTo(base_object.origin)

    def __del__(self):
        if self.enable_cuda:
            self.unregister()
        super(BaseCamera, self).__del__()

    """
    Following functions are cuda support
    """

    @property
    def cuda_array(self):
        assert self.mapped
        return cp.ndarray(
            shape=(self.cuda_shape[1], self.cuda_shape[0], 4),
            dtype=self.cuda_dtype,
            strides=self.cuda_strides,
            order=self.cuda_order,
            memptr=self._cuda_buffer
        )

    @property
    def cuda_buffer(self):
        assert self.mapped
        return self._cuda_buffer

    @property
    def graphics_resource(self):
        assert self.registered
        return self.cuda_graphics_resource

    @property
    def registered(self):
        return self.cuda_graphics_resource is not None

    @property
    def mapped(self):
        return self._cuda_buffer is not None

    def __enter__(self):
        return self.map()

    def __exit__(self, exc_type, exc_value, trace):
        self.unmap()
        return False

    def register(self):
        self.cuda_texture_identifier = self.texture_context_future.result().getNativeId()
        assert self.cuda_texture_identifier is not None
        if self.registered:
            return self.cuda_graphics_resource
        self.cuda_graphics_resource = check_cudart_err(
            cudart.cudaGraphicsGLRegisterImage(
                self.cuda_texture_identifier, GL_TEXTURE_2D, cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsReadOnly
            )
        )
        return self.cuda_graphics_resource

    def unregister(self):
        if self.registered:
            self.unmap()
            self.cuda_graphics_resource = check_cudart_err(
                cudart.cudaGraphicsUnregisterResource(self.cuda_graphics_resource)
            )
            self.cam.node().getDisplayRegion(0).clearDrawCallback()

    def map(self, stream=0):
        if not self.registered:
            raise RuntimeError("Cannot map an unregistered buffer.")
        if self.mapped:
            return self._cuda_buffer
        check_cudart_err(cudart.cudaGraphicsMapResources(1, self.cuda_graphics_resource, stream))
        array = check_cudart_err(cudart.cudaGraphicsSubResourceGetMappedArray(self.graphics_resource, 0, 0))
        channelformat, cudaextent, flag = check_cudart_err(cudart.cudaArrayGetInfo(array))

        depth = 1
        byte = 4  # four channel
        if self.new_cuda_mem_ptr is None:
            success, self.new_cuda_mem_ptr = cudart.cudaMalloc(cudaextent.height * cudaextent.width * byte * depth)
        check_cudart_err(
            cudart.cudaMemcpy2DFromArray(
                self.new_cuda_mem_ptr, cudaextent.width * byte * depth, array, 0, 0, cudaextent.width * byte * depth,
                cudaextent.height, cudart.cudaMemcpyKind.cudaMemcpyDeviceToDevice
            )
        )
        if self._cuda_buffer is None:
            self._cuda_buffer = cp.cuda.MemoryPointer(
                cp.cuda.UnownedMemory(self.new_cuda_mem_ptr, cudaextent.width * depth * byte * cudaextent.height, self),
                0
            )
        return self.cuda_array

    def unmap(self, stream=None):
        if not self.registered:
            raise RuntimeError("Cannot unmap an unregistered buffer.")
        if not self.mapped:
            return self
        self._cuda_buffer = check_cudart_err(cudart.cudaGraphicsUnmapResources(1, self.cuda_graphics_resource, stream))
        return self



from metadrive.component.algorithm.BIG import BIG
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.engine.asset_loader import initialize_asset_loader
from metadrive.tests.vis_block.vis_block_base import TestBlock


def vis_big(debug: bool = False):
    test = TestBlock(debug=debug)

    test.cam.setPos(250, 100, 2000)

    initialize_asset_loader(test)
    global_network = NodeRoadNetwork()

    big = BIG(2, 3.5, global_network, test.render, test.world, random_seed=5)
    test.vis_big(big)
    test.big.block_num = 45
    # big.generate(BigGenerateMethod.BLOCK_NUM, 10)
    test.run()


if __name__ == "__main__":
    vis_big()



import logging

import numpy as np
import tqdm

from metadrive.type import MetaDriveType

try:
    import geopandas as gpd
    from nuplan.common.actor_state.state_representation import Point2D
    from nuplan.common.maps.maps_datatypes import SemanticMapLayer, StopLineType
    from shapely.ops import unary_union
except ImportError:
    pass

from metadrive.component.lane.nuplan_lane import NuPlanLane
from metadrive.component.map.base_map import BaseMap
from metadrive.component.nuplan_block.nuplan_block import LaneLineProperty
from metadrive.component.nuplan_block.nuplan_block import NuPlanBlock
from metadrive.component.road_network.edge_road_network import EdgeRoadNetwork
from metadrive.constants import PGLineColor, PGLineType
from metadrive.engine.engine_utils import get_global_config
from metadrive.engine.scene_cull import SceneCull
from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector, metadrive_to_nuplan_vector

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class NuPlanMap(BaseMap):
    def __init__(self, map_name, nuplan_center, radius, random_seed=None, need_lane_localization=True):
        self.need_lane_localization = need_lane_localization
        self.map_name = map_name
        self._center = np.array(nuplan_center)
        self._nuplan_map_api = self.engine.data_manager.current_scenario.map_api
        self.attached_blocks = []
        self.boundary_block = None  # it won't be detached
        self._radius = radius
        self.cull_dist = get_global_config()["scenario_radius"]
        super(NuPlanMap, self).__init__(dict(id=map_name), random_seed=random_seed)

    @property
    def nuplan_center(self):
        return self._center

    def attach_to_world(self, center_point, parent_np=None, physics_world=None):
        parent_node_path, physics_world = self.engine.worldNP or parent_np, self.engine.physics_world or physics_world
        self.road_network = self.road_network_type()
        for block in self.blocks:
            # block.block_network.show_bounding_box(self.engine)
            if not SceneCull.out_of_bounding_box(block.bounding_box, np.array(center_point) - self.nuplan_center,
                                                 self.cull_dist):
                self.road_network.add(block.block_network)
                block.attach_to_world(parent_node_path, physics_world)
                self.attached_blocks.append(block)
        if not self.engine.global_config["load_city_map"]:
            self.boundary_block.attach_to_world(parent_node_path, physics_world)

    def detach_from_world(self, physics_world=None):
        if not self.engine.global_config["load_city_map"]:
            self.boundary_block.detach_from_world(self.engine.physics_world or physics_world)
        for block in self.attached_blocks:
            block.detach_from_world(self.engine.physics_world or physics_world)

    def _generate(self):
        logger.info("\n \n ############### Start Building Map: {} ############### \n".format(self.map_name))
        np.seterr(all='ignore')
        map_api = self._nuplan_map_api
        # Center is Important !
        center = self.nuplan_center
        layer_names = [
            SemanticMapLayer.LANE_CONNECTOR,
            SemanticMapLayer.LANE,
            SemanticMapLayer.CROSSWALK,
            SemanticMapLayer.INTERSECTION,
            SemanticMapLayer.STOP_LINE,
            SemanticMapLayer.WALKWAYS,
            SemanticMapLayer.CARPARK_AREA,
            SemanticMapLayer.ROADBLOCK,
            SemanticMapLayer.ROADBLOCK_CONNECTOR,

            # unsupported yet
            # SemanticMapLayer.STOP_SIGN,
            # SemanticMapLayer.DRIVABLE_AREA,
        ]
        center_for_query = Point2D(*metadrive_to_nuplan_vector(center))
        nearest_vector_map = map_api.get_proximal_map_objects(center_for_query, self._radius, layer_names)
        # Filter out stop polygons in turn stop
        if SemanticMapLayer.STOP_LINE in nearest_vector_map:
            stop_polygons = nearest_vector_map[SemanticMapLayer.STOP_LINE]
            nearest_vector_map[SemanticMapLayer.STOP_LINE] = [
                stop_polygon for stop_polygon in stop_polygons if stop_polygon.stop_line_type != StopLineType.TURN_STOP
            ]

        block_polygons = []
        # Lane and lane line
        block_index = 0
        name = {SemanticMapLayer.ROADBLOCK: "Road Block", SemanticMapLayer.ROADBLOCK_CONNECTOR: "Road Connector"}

        for layer in tqdm.tqdm([SemanticMapLayer.ROADBLOCK, SemanticMapLayer.ROADBLOCK_CONNECTOR]):
            for block in tqdm.tqdm(nearest_vector_map[layer], leave=False, desc="Building {}".format(name[layer])):
                road_block = NuPlanBlock(block_index, self.road_network, 0, self.map_name, self.nuplan_center)

                # We implement the sample() function outside the Block instance, block._sample() will do nothing
                def _sample_topology():
                    for lane_meta_data in block.interior_edges:
                        if hasattr(lane_meta_data, "baseline_path"):
                            road_block.block_network.add_lane(
                                NuPlanLane(
                                    nuplan_center=center,
                                    lane_meta_data=lane_meta_data,
                                    need_lane_localization=self.need_lane_localization
                                )
                            )
                            is_connector = True if layer == SemanticMapLayer.ROADBLOCK_CONNECTOR else False
                            road_block.set_lane_line(
                                lane_meta_data, is_road_connector=is_connector, nuplan_center=center
                            )

                if layer == SemanticMapLayer.ROADBLOCK:
                    block_polygons.append(block.polygon)
                block_index += 1
                setattr(road_block, "_sample_topology", _sample_topology)
                road_block.construct_block(self.engine.worldNP, self.engine.physics_world, attach_to_world=False)
                self.blocks.append(road_block)
                # intersection road connector

        self.boundary_block = NuPlanBlock(block_index, self.road_network, 0, self.map_name, self.nuplan_center)
        interpolygons = [block.polygon for block in nearest_vector_map[SemanticMapLayer.INTERSECTION]]
        # logger.warning("Stop using boundaries! Use exterior instead!")
        boundaries = gpd.GeoSeries(unary_union(interpolygons + block_polygons)).boundary.explode(index_parts=True)
        # boundaries.plot()
        # plt.show()
        for idx, boundary in enumerate(boundaries[0]):
            block_points = np.array(list(i for i in zip(boundary.coords.xy[0], boundary.coords.xy[1])))
            block_points = nuplan_to_metadrive_vector(block_points, self.nuplan_center)
            id = "boundary_{}".format(idx)
            self.boundary_block.lines[id] = LaneLineProperty(
                id, block_points, PGLineColor.GREY, PGLineType.CONTINUOUS, in_road_connector=False
            )
        self.boundary_block.construct_block(self.engine.worldNP, self.engine.physics_world, attach_to_world=True)
        np.seterr(all='warn')

        logger.info("\n \n ############### Finish Building Map: {} ############### \n".format(self.map_name))

    def play(self):
        # For debug
        for b in self.blocks:
            b._create_in_world(skip=True)
            b.attach_to_world(self.engine.worldNP, self.engine.physics_world)
            b.detach_from_world(self.engine.physics_world)

    def nuplan_to_metadrive_position(self, pos):
        return nuplan_to_metadrive_vector(pos, self.nuplan_center)

    @property
    def road_network_type(self):
        return EdgeRoadNetwork

    def destroy(self):
        self.map_name = None
        super(NuPlanMap, self).destroy()

    def __del__(self):
        # self.destroy()
        logging.debug("Map is Released")
        print("[NuPlanMap] Map is Released")

    def show_coordinates(self):
        lanes = [lane_info.lane for lane_info in self.road_network.graph.values()]
        self.engine.show_lane_coordinates(lanes)

    def get_boundary_line_vector(self, interval):
        ret = {}
        for block in self.attached_blocks + [self.boundary_block]:
            for boundary in block.lines.values():
                type = boundary.type
                map_feat_id = str(boundary.id)

                if type == PGLineType.BROKEN:
                    ret[map_feat_id] = {
                        "type": MetaDriveType.LINE_BROKEN_SINGLE_YELLOW
                        if boundary.color == PGLineColor.YELLOW else MetaDriveType.LINE_BROKEN_SINGLE_WHITE,
                        "polyline": boundary.points
                    }
                else:
                    ret[map_feat_id] = {
                        "polyline": boundary.points,
                        "type": MetaDriveType.LINE_SOLID_SINGLE_YELLOW
                        if boundary.color == PGLineColor.YELLOW else MetaDriveType.LINE_SOLID_SINGLE_WHITE
                    }
        return ret

    # def get_center_point(self):
    #     "Map is set to 0,0 in nuplan map"
    #     return [0, 0]


if __name__ == "__main__":
    from metadrive.envs.real_data_envs.nuplan_env import NuPlanEnv
    from metadrive.manager.nuplan_data_manager import NuPlanDataManager
    from metadrive.engine.engine_utils import initialize_engine, set_global_random_seed

    default_config = NuPlanEnv.default_config()
    default_config["use_render"] = True
    default_config["debug"] = True
    default_config["show_coordinates"] = True
    default_config["debug_static_world"] = True
    engine = initialize_engine(default_config)
    set_global_random_seed(0)

    engine.data_manager = NuPlanDataManager()
    engine.data_manager.seed(0)

    center = nuplan_to_metadrive_vector([664396, 3997613])

    map = NuPlanMap(map_name=0, nuplan_center=center, radius=500)
    map.attach_to_world(center)
    # engine.enableMouse()
    map.road_network.show_bounding_box(engine, (1, 0, 0, 1))
    lanes = [lane_info.lane for lane_info in map.road_network.graph.values()]
    engine.show_lane_coordinates(lanes)

    def detach_map():
        map.road_network.remove_bounding_box()
        map.detach_from_world()

    def attach_map():
        position = np.array(center)
        map.attach_to_world(position)
        map.road_network.show_bounding_box(engine, (1, 0, 0, 1))
        engine.main_camera.set_bird_view_pos(pos)

    engine.accept("d", detach_map)
    engine.accept("a", attach_map)

    pos = map.get_center_point()
    engine.main_camera.set_bird_view_pos(pos)

    while True:
        map.engine.step()



import os
from os.path import dirname, realpath, isdir
os.chdir(dirname(realpath(__file__)))

if not isdir("mips"):
    os.makedirs("mips")

for i in range(12):
    num_samples = 256 + i * 12
    if i == 0:
        num_samples = 128
    with open('mips/{}.autogen.glsl'.format(i), 'w') as handle:
        handle.write('// Autogenerated, do not edit\n')
        handle.write('#version 430\n')
        handle.write('#define SHADER_NUM_SAMPLES {}\n'.format(num_samples))
        handle.write('#pragma include "../filter_cubemap.frag.glsl"\n')



import copy
import logging

from metadrive.base_class.base_object import BaseObject
from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import PGMap, MapGenerateMethod
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.component.vehicle.vehicle_type import DefaultVehicle
from metadrive.constants import DEFAULT_AGENT
from metadrive.constants import ObjectState
from metadrive.constants import PolicyState
from metadrive.constants import REPLAY_DONE
from metadrive.manager.base_manager import BaseManager
from metadrive.obs.state_obs import LidarStateObservation
from metadrive.utils import recursive_equal


class ReplayManager(BaseManager):
    PRIORITY = 99  # lowest

    def __init__(self):
        super(ReplayManager, self).__init__()
        self.restore_episode_info = None
        self.current_map = None
        self.current_frames = None
        self.current_frame = None
        self.replay_done = False
        self.record_name_to_current_name = dict()
        self.current_name_to_record_name = dict()
        self.observation = self.get_observation()

    def before_reset(self, *args, **kwargs):
        """
        Clean generated objects
        """
        if self.engine.only_reset_when_replay:
            for name in self.spawned_objects.keys():
                assert name not in self.engine._spawned_objects, \
                    "Other Managers failed to clean objects loaded by ReplayManager"
            self.spawned_objects = {}
            assert len(self.engine._object_policies) == 0, "Policy should be cleaned for reducing memory usage"
        else:
            self.clear_objects([name for name in self.spawned_objects])
        self.replay_done = False

    def spawn_object(self, object_class, **kwargs):
        """
        Spawn one objects
        """
        object = self.engine.spawn_object(object_class, **kwargs)
        if not isinstance(object, BaseMap):
            # map has different treatment as what has been done in MapManager
            self.spawned_objects[object.id] = object
        return object

    def reset(self):
        if not self.engine.replay_episode:
            return
        assert not self.engine.record_episode, "When replay, please set record to False"
        self.record_name_to_current_name = dict()
        self.current_name_to_record_name = dict()
        self.restore_episode_info = self.engine.global_config["replay_episode"]
        self.restore_episode_info["frame"].reverse()
        # Since in episode data map data only contains one map, values()[0] is the map_parameters
        map_data = self.restore_episode_info["map_data"]
        assert len(map_data) > 0, "Can not find map info in episode data"

        map_config = copy.deepcopy(map_data["map_config"])
        map_config[BaseMap.GENERATE_TYPE] = MapGenerateMethod.PG_MAP_FILE
        map_config[BaseMap.GENERATE_CONFIG] = map_data["block_sequence"]
        if self.engine.map_manager.maps[self.engine.global_seed] is not None:
            self.current_map = self.engine.map_manager.maps[self.engine.global_seed]
            assert recursive_equal(
                self.current_map.get_meta_data()["block_sequence"], map_data["block_sequence"], need_assert=True
            ), "Loaded data mismatch stored data"
            self.engine.map_manager.load_map(self.current_map)
        else:
            self.current_map = self.spawn_object(
                PGMap, map_config=map_config, auto_fill_random_seed=False, force_spawn=True
            )
        self.current_frames = self.restore_episode_info["frame"].pop()
        self.replay_frame()
        if self.engine.only_reset_when_replay:
            # do not replay full trajectory! set state for managers for interaction
            self.restore_manager_states(self.current_frame.manager_info)
            # Do special treatment to map manager
            self.engine.map_manager.current_map = self.current_map
            self.engine.map_manager.maps[self.engine.global_seed] = self.current_map

    def restore_policy_states(self, policy_spawn_infos):
        # restore agent policy
        agent_policy = self.engine.agent_manager.agent_policy
        agent_obj_name = self.engine.agent_manager.active_agents[DEFAULT_AGENT].name
        for name, policy_spawn_info in policy_spawn_infos.items():
            obj_name = self.record_name_to_current_name[name]
            p_class = policy_spawn_info[PolicyState.POLICY_CLASS] if obj_name != agent_obj_name else agent_policy
            args = policy_spawn_info[PolicyState.ARGS]
            kwargs = policy_spawn_info[PolicyState.KWARGS]
            assert obj_name == self.record_name_to_current_name[policy_spawn_info[PolicyState.OBJ_NAME]]
            assert obj_name in self.engine.get_objects().keys(), "Can not find obj when restoring policies"
            policy = self.add_policy(obj_name, p_class, *args, **kwargs)
            if policy.control_object is BaseObject:
                obj = list(self.engine.get_objects([obj_name]).values())[0]
                policy.control_object = obj
                assert obj.id == obj_name

    def restore_manager_states(self, states):
        current_managers = [manager.class_name for manager in self.engine.managers.values()]
        data_managers = states.keys()
        assert len(current_managers - data_managers
                   ) == 0, "Manager not match, data: {}, current: {}".format(data_managers, current_managers)
        for manager in self.engine.managers.values():
            manager.set_state(states[manager.class_name], old_name_to_current=self.record_name_to_current_name)

    def step(self, *args, **kwargs):
        # Note: Update object state must be written in step, because the simulator will step 5 times for each RL step.
        # We need to record the intermediate states.
        if self.engine.replay_episode and not self.engine.only_reset_when_replay:
            self.replay_frame()

    def after_step(self, *args, **kwargs):
        if self.engine.replay_episode and not self.engine.only_reset_when_replay:
            if len(self.restore_episode_info["frame"]) == 0:
                self.replay_done = True
            return self.engine.agent_manager.for_each_active_agents(lambda v: {REPLAY_DONE: self.replay_done})
        else:
            return dict()

    def destroy(self):
        self.record_name_to_current_name = dict()
        self.current_name_to_record_name = dict()
        self.restore_episode_info = None
        self.current_map = None

    def replay_frame(self):
        if len(self.current_frames) == 0:
            return
        self.current_frame = self.current_frames.pop()
        # create
        for name, config in self.current_frame.spawn_info.items():
            if config[ObjectState.CLASS] == DefaultVehicle:
                config[ObjectState.INIT_KWARGS]["vehicle_config"]["use_special_color"] = True
            obj = self.spawn_object(object_class=config[ObjectState.CLASS], **config[ObjectState.INIT_KWARGS])
            self.current_name_to_record_name[obj.name] = name
            self.record_name_to_current_name[name] = obj.name
            if issubclass(config[ObjectState.CLASS], BaseVehicle):
                obj.navigation.set_route(
                    self.current_frame.step_info[name]["spawn_road"],
                    self.current_frame.step_info[name]["destination"][-1]
                )
        if self.engine.only_reset_when_replay:
            # for generation policies
            self.restore_policy_states(self.current_frame.policy_spawn_info)
        else:
            # Do not set position, in this mode, or randomness will be introduced!
            for name, state in self.current_frame.step_info.items():
                self.spawned_objects[self.record_name_to_current_name[name]].before_step()
                self.spawned_objects[self.record_name_to_current_name[name]].set_state(state)
                self.spawned_objects[self.record_name_to_current_name[name]].after_step()

        to_clear = []
        for name in self.current_frame.clear_info:
            to_clear.append(self.record_name_to_current_name[name])
            self.current_name_to_record_name.pop(self.record_name_to_current_name[name])
            self.record_name_to_current_name.pop(name)
        self.clear_objects(to_clear)

        self.replay_done = False

    @property
    def replay_agents(self):
        return {k: self.get_object_from_agent(k) for k in self.current_frame.agents}

    def __del__(self):
        logging.debug("Replay system is destroyed")

    def get_object_from_agent(self, agent_id):
        return self.spawned_objects[self.record_name_to_current_name[self.current_frame.agent_to_object(agent_id)]]

    def get_observation(self):
        """
        Override me in the future for collecting other modality
        """
        return LidarStateObservation(self.engine.global_config)

    def get_replay_agent_observations(self):
        return {k: self.observation for k in self.current_frame.agents}

    def set_state(self, state: dict, old_name_to_current=None):
        return {}

    def get_state(self):
        return {}

    def before_step(self, *args, **kwargs) -> dict:
        super(ReplayManager, self).before_step()
        if self.engine.replay_episode and not self.engine.only_reset_when_replay:
            self.current_frames = self.restore_episode_info["frame"].pop()
            self.current_frames.reverse()
        return {}



"""
This script is used to remove the optimizer state in the checkpoint. So that we can compress 2/3 of the checkpoint size.
This script is put here for reference only. In formal release, the original checkpoint file will be removed so
this script will become not runnable.
"""
import os.path as osp
import pickle

import numpy as np

ckpt_path = osp.join(osp.dirname(__file__), "checkpoint_417/checkpoint-417")
if __name__ == '__main__':
    remove_value_network = True
    path = "expert_weights.npz"

    with open(ckpt_path, "rb") as f:
        data = f.read()
    unpickled = pickle.loads(data)
    worker = pickle.loads(unpickled.pop("worker"))
    if "_optimizer_variables" in worker["state"]["default_policy"]:
        worker["state"]["default_policy"].pop("_optimizer_variables")
    pickled_worker = pickle.dumps(worker)
    weights = worker["state"]["default_policy"]
    if remove_value_network:
        weights = {k: v for k, v in weights.items() if "value" not in k}
    np.savez_compressed(path, **weights)
    print("Numpy agent weight is saved at: {}!".format(path))



"""
This script demonstrates how to setup the Safe RL environments.

Please feel free to run this script to enjoy a journey by keyboard! Remember to press H to see help message!

Auto-Drive mode may fail to solve some scenarios due to distribution mismatch.
"""
import logging

from metadrive.constants import HELP_MESSAGE
from metadrive.tests.test_functionality.test_object_collision_detection import ComplexEnv

if __name__ == "__main__":
    env = ComplexEnv(dict(use_render=True, manual_control=True, vehicle_config={"show_navi_mark": False}))
    try:
        env.reset()
        print(HELP_MESSAGE)
        env.vehicle.expert_takeover = True
        for i in range(1, 1000000000):
            previous_takeover = env.current_track_vehicle.expert_takeover
            o, r, tm, tc, info = env.step([0, 0])
            env.render(
                text={
                    "Auto-Drive (Switch mode: T)": "on" if env.current_track_vehicle.expert_takeover else "off",
                    "Total episode cost": env.episode_cost,
                    "Keyboard Control": "W,A,S,D",
                }
            )
            if not previous_takeover and env.current_track_vehicle.expert_takeover:
                logging.warning("Auto-Drive mode may fail to solve some scenarios due to distribution mismatch")
            if (tm or tc) and info["arrive_dest"]:
                env.reset()
                env.current_track_vehicle.expert_takeover = True
    except Exception as e:
        raise e
    finally:
        env.close()



import copy
from metadrive.utils.utils import get_time_str
import logging

from metadrive.base_class.base_object import BaseObject
from metadrive.constants import ObjectState, PolicyState
from metadrive.manager.base_manager import BaseManager
from metadrive.utils.utils import is_map_related_instance, is_map_related_class


class FrameInfo:
    def __init__(self, episode_step):
        self.episode_step = episode_step
        # used to track the objects spawn info
        self.spawn_info = {}
        self.policy_info = {}
        self.policy_spawn_info = {}
        # used to track update objects state in the scene
        self.step_info = {}
        # used to track the objects cleared
        self.clear_info = []
        # manager state
        self.manager_info = {}
        self.agents = []
        self._object_to_agent = None
        self._agent_to_object = None

    def object_to_agent(self, object):
        return self._object_to_agent[object]

    def agent_to_object(self, agent):
        return self._agent_to_object[agent]


class RecordManager(BaseManager):
    """
    Record the episode information for replay or reloading episode
    """
    PRIORITY = 100  # lowest priority

    def __init__(self):
        super(RecordManager, self).__init__()
        self.episode_info = None
        self.current_frames = None
        self.current_frame_count = 0
        self.reset_frame = None
        # for debug, we don't allow assign the same id to different vehicles
        # previous recycling mechanism will bring such issue, which is fixed now
        self._episode_obj_names = set()

    def before_reset(self):
        if self.engine.record_episode:
            self.episode_info = {}
            self._episode_obj_names = set()
            self.reset_frame = FrameInfo(self.engine.episode_step)

    def after_reset(self):
        """
        create a new log to record, note: after_step will be called after calling after_reset()
        """
        if self.engine.record_episode:
            self.episode_info = dict(
                map_data=self.engine.current_map.get_meta_data(),
                frame=[[self.reset_frame]],
                scenario_index=self.engine.global_seed,
                global_config=self.engine.global_config,
                global_seed=self.engine.global_seed,
                manager_metadata={},
                coordinate="MetaDrive",
                time=get_time_str()
            )

            self.collect_objects_states()
            self.collect_manager_states()
            self.collect_manager_metadata()
            self.current_frames = None
            self.reset_frame = None
            self.current_frame_count = 0

    def collect_manager_metadata(self):
        assert self.episode_step == 0, "This func can only be called after env.reset() without any env.step() called"
        ret = {}
        for manager in self.engine.managers.values():
            mgr_meta = manager.get_metadata()
            assert mgr_meta is not None, "No return value for manager.get_state()"
            ret[manager.class_name] = mgr_meta
        self.episode_info["manager_metadata"] = ret

    def collect_manager_states(self):
        ret = {}
        for manager in self.engine.managers.values():
            mgr_state = manager.get_state()
            assert mgr_state is not None, "No return value for manager.get_state()"
            ret[manager.class_name] = mgr_state
        self.current_frame.manager_info = ret

    def before_step(self, *args, **kwargs) -> dict:
        if self.engine.record_episode:
            self.current_frames = [
                FrameInfo(self.engine.episode_step) for _ in range(self.engine.global_config["decision_repeat"])
            ]
            self.current_frame_count = 0
        return {}

    def step(self, *args, **kwargs):
        # Note: Update object state must be written in step, because the simulator will step 5 times for each RL step.
        # We need to record the intermediate states.
        if self.engine.record_episode:
            self.collect_objects_states()
            self.collect_manager_states()
            self.current_frame_count += 1 if self.current_frame_count < len(self.current_frames) - 1 else 0

    def after_step(self, *args, **kwargs) -> dict:
        # frame count ==0 is the reset frame, so don't append
        if self.engine.record_episode and self.current_frame_count:
            self.step()
            assert len(self.current_frames) == self.engine.global_config["decision_repeat"], "Number of Frame Mismatch!"
            self.episode_info["frame"].append(self.current_frames)
        return {}

    def collect_objects_states(self):
        policy_mapping = self.engine.get_policies()
        for name, obj in self.engine.get_objects().items():
            if not is_map_related_instance(obj):
                self.current_frame.step_info[name] = obj.get_state()
                if name in policy_mapping:
                    self.current_frame.policy_info[name] = policy_mapping[name].get_state()

        self.current_frame.agents = list(self.engine.agents.keys())
        self.current_frame._agent_to_object = copy.deepcopy(self.engine.agent_manager._agent_to_object)
        self.current_frame._object_to_agent = copy.deepcopy(self.engine.agent_manager._object_to_agent)

    def get_episode_metadata(self):
        assert self.engine.record_episode, "Turn on recording episode and then dump it"
        return copy.deepcopy(self.episode_info)

    def destroy(self):
        self.episode_info = None

    def add_spawn_info(self, obj, object_class, kwargs):
        """
        Call when spawn new objects, ignore map related things
        """
        if not is_map_related_class(object_class) and self.engine.record_episode:
            name = obj.name
            assert name not in self.current_frame.spawn_info, "Duplicated record!"
            assert name not in self._episode_obj_names, "Duplicated name using!"
            self._episode_obj_names.add(name)
            self.current_frame.spawn_info[name] = {
                ObjectState.CLASS: object_class,
                ObjectState.INIT_KWARGS: kwargs,
                ObjectState.NAME: name
            }

            # update step info in after step so exempt the requirement for adding info here
            # self.current_frame.step_info[name] = obj.get_state()
            # policy_mapping = self.engine.get_policies()
            # if name in policy_mapping:
            #     self.current_frame.policy_info[name] = policy_mapping[name].get_state()

    def add_policy_info(self, name, policy_class, *args, **kwargs):
        """
        Call when spawn new objects, ignore map related stuff
        """
        filtered_args = []
        for arg in args:
            filtered_args.append(arg) if not isinstance(arg, BaseObject) else filtered_args.append(BaseObject)
        filtered_kwargs = {}
        for k, v in kwargs.items():
            filtered_kwargs[k] = v if not isinstance(v, BaseObject) else BaseObject
        if self.engine.record_episode:
            assert name not in self.current_frame.policy_spawn_info, "Duplicated record!"
            self.current_frame.policy_spawn_info[name] = {
                PolicyState.POLICY_CLASS: policy_class,
                PolicyState.ARGS: filtered_args,
                PolicyState.KWARGS: filtered_kwargs,
                PolicyState.OBJ_NAME: name
            }

    def add_clear_info(self, obj):
        """
        Call when clear objects, ignore map related things
        """
        if not is_map_related_instance(obj) and self.engine.record_episode and self.episode_step != 0:
            self.current_frame.clear_info.append(obj.name)

    def __del__(self):
        logging.debug("Record system is destroyed")

    @property
    def current_frame(self):
        return self.current_frames[self.current_frame_count] if self.reset_frame is None else self.reset_frame

    def set_state(self, state: dict, old_name_to_current=None):
        return {}

    def get_state(self):
        return {}



import time

from metadrive import MetaDriveEnv
from metadrive.utils import setup_logger


def _test_fps():
    # print("Start to profile the efficiency of MetaDrive with 1000 maps and ~8 vehicles!")
    try:
        setup_logger(debug=False)
        env = MetaDriveEnv(dict(
            num_scenarios=1000,
            start_seed=1010,
        ))
        obs, _ = env.reset()
        start = time.time()
        action = [0.0, 1.]
        total_steps = 5000
        vehicle_num = [len(env.engine.traffic_manager.vehicles)]
        for s in range(total_steps):
            o, r, tm, tc, i = env.step(action)
            if tm or tc:
                env.reset()
                vehicle_num.append(len(env.engine.traffic_manager.vehicles))
        assert total_steps / (time.time() - start) > 200
    finally:
        env.close()


if __name__ == '__main__':
    _test_fps()



import os
import pickle
import shutil

import numpy as np

from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.envs.real_data_envs.nuscenes_env import NuScenesEnv
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.policy.replay_policy import ReplayEgoCarPolicy
from metadrive.scenario.utils import NP_ARRAY_DECIMAL
from metadrive.scenario.utils import assert_scenario_equal
from metadrive.scenario.utils import read_dataset_summary
from metadrive.type import MetaDriveType
from metadrive.utils.math import wrap_to_pi


def test_export_metadrive_scenario_reproduction(num_scenarios=3, render_export_env=False, render_load_env=False):
    env = MetaDriveEnv(
        dict(start_seed=0, use_render=render_export_env, num_scenarios=num_scenarios, agent_policy=IDMPolicy)
    )
    policy = lambda x: [0, 1]
    dir1 = None
    try:
        scenarios, done_info = env.export_scenarios(policy, scenario_index=[i for i in range(num_scenarios)])
        dir1 = os.path.join(os.path.dirname(__file__), "../test_component/test_export")
        os.makedirs(dir1, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir1, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()

    # Same environment, same config
    env = MetaDriveEnv(
        dict(start_seed=0, use_render=render_load_env, num_scenarios=num_scenarios, agent_policy=IDMPolicy)
    )
    policy = lambda x: [0, 1]
    try:
        scenarios2, done_info = env.export_scenarios(policy, scenario_index=[i for i in range(num_scenarios)])
    finally:
        env.close()

    if dir1 is not None:
        shutil.rmtree(dir1)

    # We can't make sure traffic vehicles has same name, so just test SDC here.
    assert_scenario_equal(scenarios, scenarios2, only_compare_sdc=True)


def test_export_metadrive_scenario_easy(num_scenarios=5, render_export_env=False, render_load_env=False):
    # ===== Save data =====
    env = MetaDriveEnv(
        dict(
            start_seed=0, map="SCS", use_render=render_export_env, num_scenarios=num_scenarios, agent_policy=IDMPolicy
        )
    )
    policy = lambda x: [0, 1]
    dir1 = None
    try:
        scenarios, done_info = env.export_scenarios(policy, scenario_index=[i for i in range(num_scenarios)])
        dir1 = os.path.join(os.path.dirname(__file__), "test_export_scenario_consistancy")
        os.makedirs(dir1, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir1, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()
        # pass

    # ===== Save data of the restoring environment =====
    env = WaymoEnv(
        dict(
            agent_policy=ReplayEgoCarPolicy,
            data_directory=dir1,
            use_render=render_load_env,
            num_scenarios=num_scenarios,
            force_reuse_object_name=True,
            horizon=1000,
            # debug=True,
            # debug_static_world=True,
            vehicle_config=dict(no_wheel_friction=True)
        )
    )
    try:
        scenarios_restored, done_info = env.export_scenarios(
            policy,
            scenario_index=[i for i in range(num_scenarios)],
            render_topdown=render_load_env,
            return_done_info=True
        )
        for seed, info in done_info.items():
            if not info["arrive_dest"]:
                raise ValueError("Seed: {} Can not arrive dest!".format(seed))
    finally:
        env.close()

    if dir1 is not None:
        shutil.rmtree(dir1)

    assert_scenario_equal(scenarios, scenarios_restored, only_compare_sdc=False)


def test_export_metadrive_scenario_hard(start_seed=0, num_scenarios=3, render_export_env=False, render_load_env=False):
    # ===== Save data =====
    env = MetaDriveEnv(
        dict(
            start_seed=start_seed,
            map=7,
            use_render=render_export_env,
            num_scenarios=num_scenarios,
            agent_policy=IDMPolicy
        )
    )
    policy = lambda x: [0, 1]
    dir1 = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(start_seed, start_seed + num_scenarios)]
        )
        dir1 = os.path.join(os.path.dirname(__file__), "test_export_metadrive_scenario_hard")
        os.makedirs(dir1, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir1, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()
        # pass

    # ===== Save data of the restoring environment =====
    env = WaymoEnv(
        dict(
            agent_policy=ReplayEgoCarPolicy,
            data_directory=dir1,
            use_render=render_load_env,
            num_scenarios=num_scenarios,
            start_scenario_index=start_seed,
            debug=True,
            force_reuse_object_name=True,
            vehicle_config=dict(no_wheel_friction=True)
            # debug_physics_world=True,
            # debug_static_world=True
        )
    )
    try:
        scenarios_restored, done_info = env.export_scenarios(
            policy,
            scenario_index=[i for i in range(num_scenarios)],
            render_topdown=render_load_env,
            return_done_info=True
        )
        for seed, info in done_info.items():
            if not info["arrive_dest"]:
                raise ValueError("Seed: {} Can not arrive dest!".format(seed))
    finally:
        env.close()

    if dir1 is not None:
        shutil.rmtree(dir1)

    assert_scenario_equal(scenarios, scenarios_restored, only_compare_sdc=False)


def test_export_waymo_scenario(num_scenarios=3, render_export_env=False, render_load_env=False):
    env = WaymoEnv(
        dict(
            agent_policy=ReplayEgoCarPolicy,
            use_render=render_export_env,
            start_scenario_index=0,
            num_scenarios=num_scenarios
        )
    )
    policy = lambda x: [0, 1]
    dir = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(num_scenarios)], verbose=True
        )
        dir = os.path.join(os.path.dirname(__file__), "../test_component/test_export")
        os.makedirs(dir, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()

    try:
        print("===== Start restoring =====")
        env = WaymoEnv(
            dict(
                agent_policy=ReplayEgoCarPolicy,
                data_directory=dir,
                use_render=render_load_env,
                num_scenarios=num_scenarios,
                force_reuse_object_name=True,
                vehicle_config=dict(no_wheel_friction=True)
            )
        )
        scenarios_restored, done_info = env.export_scenarios(
            policy,
            scenario_index=[i for i in range(num_scenarios)],
            render_topdown=render_load_env,
            return_done_info=True
        )
        for seed, info in done_info.items():
            if not info["arrive_dest"]:
                raise ValueError("Seed: {} Can not arrive dest!".format(seed))

    finally:
        env.close()
        # if dir is not None:
        #     shutil.rmtree(dir)


def test_export_nuscenes_scenario(num_scenarios=2, render_export_env=False, render_load_env=False):
    env = NuScenesEnv(
        dict(
            data_directory=AssetLoader.file_path("nuscenes", return_raw_style=False),
            agent_policy=ReplayEgoCarPolicy,
            use_render=render_export_env,
            start_scenario_index=0,
            num_scenarios=num_scenarios
        )
    )
    policy = lambda x: [0, 1]
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(num_scenarios)], verbose=True
        )
        dir = os.path.join(os.path.dirname(__file__), "../test_component/test_export")
        os.makedirs(dir, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()

    try:
        print("===== Start restoring =====")
        env = NuScenesEnv(
            dict(
                agent_policy=ReplayEgoCarPolicy,
                data_directory=dir,
                use_render=render_load_env,
                num_scenarios=num_scenarios,
                force_reuse_object_name=True,
                vehicle_config=dict(no_wheel_friction=True)
            )
        )
        scenarios_restored, done_info = env.export_scenarios(
            policy,
            scenario_index=[i for i in range(num_scenarios)],
            render_topdown=render_load_env,
            return_done_info=True
        )
        for seed, info in done_info.items():
            if not info["arrive_dest"]:
                raise ValueError("Seed: {} Can not arrive dest!".format(seed))
        if os.path.exists(dir):
            shutil.rmtree(dir)
    finally:
        env.close()
        print("Finish!")
        # if dir is not None:
        #     shutil.rmtree(dir)


def compare_exported_scenario_with_origin(scenarios, data_manager, data_dir="waymo"):
    _, _, mapping = read_dataset_summary(AssetLoader.file_path(data_dir, return_raw_style=False))
    for index, scenario in scenarios.items():
        file_name = data_manager.summary_lookup[index]
        file_path = AssetLoader.file_path(data_dir, mapping[file_name], file_name, return_raw_style=False)
        with open(file_path, "rb+") as file:
            origin_data = pickle.load(file)
        export_data = scenario
        new_tracks = export_data["tracks"]
        original_ids = [new_tracks[obj_name]["metadata"]["original_id"] for obj_name in new_tracks.keys()]
        # assert len(set(original_ids)) == len(origin_data["tracks"]), "Object Num mismatch!"
        for obj_id, track in new_tracks.items():
            # if obj_id != scenario["metadata"]["sdc_id"]:
            new_pos = track["state"]["position"]
            new_heading = track["state"]["heading"]
            new_valid = track["state"]["valid"]
            old_id = track["metadata"]["original_id"]
            old_track = origin_data["tracks"][old_id]
            old_pos = old_track["state"]["position"]
            old_heading = old_track["state"]["heading"]
            old_valid = old_track["state"]["valid"]

            if track["type"] in [MetaDriveType.TRAFFIC_BARRIER, MetaDriveType.TRAFFIC_CONE]:
                index_to_compare = np.where(old_valid[:len(new_valid)])[0]
                assert new_valid[index_to_compare].all(), "Frame mismatch!"
                decimal = 0
            else:
                index_to_compare = np.where(new_valid)[0]
                decimal = NP_ARRAY_DECIMAL
            assert old_valid[index_to_compare].all(), "Frame mismatch!"
            old_pos = old_pos[index_to_compare][..., :2]
            new_pos = new_pos[index_to_compare][..., :2]
            np.testing.assert_almost_equal(old_pos, new_pos, decimal=decimal)

            old_heading = wrap_to_pi(old_heading[index_to_compare].reshape(-1))
            new_heading = wrap_to_pi(new_heading[index_to_compare].reshape(-1))
            np.testing.assert_almost_equal(old_heading, new_heading, decimal=decimal)

        for light_id, old_light in origin_data["dynamic_map_states"].items():
            new_light = export_data["dynamic_map_states"][light_id]

            if "stop_point" in old_light["state"]:
                old_pos = old_light["state"]["stop_point"][..., :2]
                old_pos = old_pos[np.where(old_pos > 0)[0][0]]

            else:
                old_pos = old_light["stop_point"]

            if "stop_point" in new_light["state"]:
                new_pos = new_light["state"]["stop_point"]
                ind = np.where(new_pos > 0)[0]
                if ind.size > 0:
                    ind = ind[0]
                else:
                    ind = 0
                new_pos = new_pos[ind]
            else:
                new_pos = new_light["stop_point"]

            length = min(len(old_pos), len(new_pos))
            np.testing.assert_almost_equal(old_pos[:2], new_pos[:2], decimal=NP_ARRAY_DECIMAL)

            if "lane" in old_light["state"]:
                old_light_lane = str(max(old_light["state"]["lane"]))
            else:
                old_light_lane = old_light["lane"]

            if "lane" in new_light["state"]:
                new_light_lane = max(new_light["state"]["lane"])
            else:
                new_light_lane = new_light["lane"]

            assert str(old_light_lane) == str(new_light_lane)

            for k, light_status in enumerate(old_light["state"]["object_state"][:length]):
                assert MetaDriveType.parse_light_status(light_status, simplifying=True) == \
                       new_light["state"]["object_state"][k]

        print("Finish Seed: {}".format(index))


def test_waymo_export_and_original_consistency(num_scenarios=3, render_export_env=False):
    env = WaymoEnv(
        dict(
            agent_policy=ReplayEgoCarPolicy,
            use_render=render_export_env,
            start_scenario_index=0,
            num_scenarios=num_scenarios,
            # force_reuse_object_name=True, # Don't allow discontinuous trajectory in our system
        )
    )
    policy = lambda x: [0, 1]
    dir = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(num_scenarios)], verbose=True
        )
        compare_exported_scenario_with_origin(scenarios, env.engine.data_manager)
    finally:
        env.close()


def test_nuscenes_export_and_original_consistency(num_scenarios=7, render_export_env=False):
    assert num_scenarios <= 7
    env = NuScenesEnv(
        dict(
            data_directory=AssetLoader.file_path("nuscenes", return_raw_style=False),
            agent_policy=ReplayEgoCarPolicy,
            use_render=render_export_env,
            start_scenario_index=3,
            num_scenarios=num_scenarios
        )
    )
    policy = lambda x: [0, 1]
    dir = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(3, 3 + num_scenarios)], verbose=True
        )
        compare_exported_scenario_with_origin(scenarios, env.engine.data_manager, data_dir="nuscenes")
    finally:
        env.close()


if __name__ == "__main__":
    # test_export_metadrive_scenario_reproduction(num_scenarios=10)
    test_export_metadrive_scenario_easy(render_export_env=False, render_load_env=False)
    # test_export_metadrive_scenario_hard(num_scenarios=3, render_export_env=True, render_load_env=True)
    # test_export_waymo_scenario(num_scenarios=3, render_export_env=False, render_load_env=False)
    # test_waymo_export_and_original_consistency(num_scenarios=3, render_export_env=False)
    # test_export_nuscenes_scenario(num_scenarios=2, render_export_env=False, render_load_env=False)
    # test_nuscenes_export_and_original_consistency()



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from . import spec


class TextureMixin(object):
    def get_images(self, material, shader):
        if self._render_type == 'rp':  # custom texture order for RP
            return (
                # p3d_Texture0 - baseColorTexture - RP Diffuse
                (('pbrMetallicRoughness', 'baseColorTexture'), self.get_diffuse(material, shader)),

                # p3d_Texture1 - metallicRoughnessTexture - RP Normal Map
                (('pbrMetallicRoughness', 'metallicRoughnessTexture'), self.get_normal_map(material, shader)),

                # p3d_Texture2 - normalTexture - RP IOR
                ('normalTexture', self.get_specular_map(material, shader)),

                # p3d_Texture3 - emissiveTexture - RP Roughness
                ('emissiveTexture', self.get_roughness_map(material, shader)),

                # put a placeholder, because all textures are required
                (None, ''),
            )
        else:
            return (
                (('pbrMetallicRoughness', 'baseColorTexture'), self.get_diffuse(material, shader)),
                (('pbrMetallicRoughness', 'metallicRoughnessTexture'), self.get_roughness_map(material, shader)),
                ('normalTexture', self.get_normal_map(material, shader)),
                ('emissiveTexture', self.get_emission_map(material, shader)),
                (None, ''),  # put a placeholder, because all textures are required
            )

    def make_texture(self, type_, image_texture):
        filepath = image_texture.image.filepath.lstrip('/')
        path = os.path.dirname(self._output)

        gltf_sampler = {
            'name': image_texture.image.name,
            'wrapS': spec.REPEAT,
            'wrapT': spec.REPEAT,
        }

        gltf_image = {
            'name': image_texture.image.name,
            'mimeType': 'image/{}'.format(image_texture.image.file_format.lower()),
        }

        if self._output.endswith('.gltf'):  # external texture
            gltf_image['uri'] = os.path.join(path, filepath)
        else:  # embedded texture
            gltf_image['extras'] = {}
            if image_texture.image.packed_file:
                gltf_image['extras']['data'] = image_texture.image.packed_file.data
            else:
                gltf_image['extras']['uri'] = os.path.join(self.get_cwd(), filepath)

        if image_texture.extension == 'CLIP':
            gltf_sampler['wrapS'] = spec.CLAMP_TO_EDGE
            gltf_sampler['wrapT'] = spec.CLAMP_TO_EDGE
        elif image_texture.extension == 'REPEAT':
            gltf_sampler['wrapS'] = spec.REPEAT
            gltf_sampler['wrapT'] = spec.REPEAT

        return gltf_sampler, gltf_image

    def make_empty_texture(self, type_):
        filepath = 'textures/unknown.png'

        if self._render_type == 'rp':
            if type(type_) == tuple:
                if 'baseColorTexture' in type_:
                    filepath = 'textures/diffuse.png'
                elif 'metallicRoughnessTexture' in type_:
                    filepath = 'textures/normal.png'
            elif type_ == 'normalTexture':
                filepath = 'textures/specular.png'
            elif type_ == 'emissiveTexture':
                filepath = 'textures/roughness.png'
        else:
            if type(type_) == tuple:
                if 'baseColorTexture' in type_:
                    filepath = 'textures/diffuse.png'
                elif 'metallicRoughnessTexture' in type_:
                    filepath = 'textures/metallic_roughness.png'
            elif type_ == 'normalTexture':
                filepath = 'textures/normal.png'
            elif type_ == 'emissiveTexture':
                filepath = 'textures/emissive.png'

        path = os.path.dirname(self._output)

        gltf_sampler = {
            'name': os.path.basename(filepath),
            'wrapS': spec.CLAMP_TO_EDGE,
            'wrapT': spec.CLAMP_TO_EDGE,
        }

        gltf_image = {
            'name': os.path.basename(filepath),
            'mimeType': 'image/png',
        }

        # gltf_image['uri'] = os.path.join(path, filepath)
        if self._output.endswith('.gltf'):  # external texture
            gltf_image['uri'] = os.path.join(path, filepath)
        else:  # embedded texture
            gltf_image['extras'] = {
                'uri': os.path.join(self.get_cwd(), filepath),
            }

        return gltf_sampler, gltf_image
